From Coq Require Import ZArith.
From mathcomp Require Import all_ssreflect.

Require Import
  syntax
  semantics
  security
  typesystem
  utils
  utils_facts
.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Open Scope Z.


(* -------------------------------------------------------------------------- *)
(* Syntax. *)
(* -------------------------------------------------------------------------- *)

Variant annot :=
  | N
  | C
  | R.

Inductive linstr :=
  | Lassign     of annot  & regvar & expression
  | Lload       of regvar & arrvar & expression
  | Lstore      of arrvar & expression & regvar
  | Linit_msf   of regvar
  | Lupdate_msf of annot  & regvar & regvar & expression
  | Lprotect    of regvar & regvar & regvar
  | Ljump       of annot
  | Lcjump      of annot  & expression.

Definition lprogram := label -> linstr * list label.

(* Semantics *)

Record lstate :=
  {
    l_pc : label;
    l_dt : data;
  }.

Definition with_pc (st : lstate) (l : label) : lstate :=
  {|
    l_pc := l ;
    l_dt := l_dt st;
  |}.

Definition with_dt (st : lstate) (dt : data) : lstate :=
  {|
    l_pc := l_pc st;
    l_dt := dt;
  |}.


(* -------------------------------------------------------------------------- *)
(* Small step semantics. *)

Definition l_after_assign st l x e :=
  let dt := l_dt st in
  let v := eval (dt_vm dt) e in
  let dt' := write_reg dt x v in
  with_pc (with_dt st dt') l.

Definition l_after_init_msf st l msf :=
  let dt' := write_reg (l_dt st) msf msf_nomask in
  with_pc (with_dt st dt') l.

Definition l_after_update_msf st l x y e :=
  let dt' := dt_after_update_msf (l_dt st) x y e in
  with_pc (with_dt st dt') l.

Definition l_after_protect st l x y msf :=
  let dt' := dt_after_protect (l_dt st) x y msf in
  with_pc (with_dt st dt') l.

Definition l_after_load st l x a i :=
  let dt := l_dt st in
  let v := Mem.read (dt_mem dt) a i in
  let dt' := write_reg dt x v in
  with_pc (with_dt st dt') l.

Definition l_after_store st l a i x :=
  let dt := l_dt st in
  let v := Vm.get (dt_vm dt) x in
  let dt' := write_mem dt a i v in
  with_pc (with_dt st dt') l.

(* The rules for unsafe loads and stores require that the misspeculation status
   is set, because we assume that the program is sequentially safe, and
   therefore unsafe memory accesses must happen under misspeculation. *)

Definition upd_ms (dt:data) (b:bool) : data :=
  with_ms dt (b || dt_ms dt).

Variant lstep
  (p : lprogram) : lstate -> directive -> observation -> lstate -> Prop :=
  | lstep_assign :
    forall a st x e l',
      p (l_pc st) = (Lassign a x e, [::l'])
      -> lstep p st Dstep Onone (l_after_assign st l' x e)

  | lstep_load_n :
    forall st x a e l' i,
      let: dt := l_dt st in
      let: st' := l_after_load st l' x a i in
      p (l_pc st) = (Lload x a e, [::l'])
      -> eval (dt_vm dt) e = Vint i
      -> in_bounds a i
      -> lstep p st Dstep (Oaddr a i) st'

  | lstep_load_s :
    forall st x a e l' i b j,
      let: dt := l_dt st in
      let: st' := l_after_load st l' x b j in
      p (l_pc st) = (Lload x a e, [::l'])
      -> eval (dt_vm dt) e = Vint i
      -> ~~ in_bounds a i
      -> in_bounds b j
      -> dt_ms dt
      -> lstep p st (Dmem b j) (Oaddr a i) st'

  | lstep_store_n :
    forall st a e x l' i,
      let: dt := l_dt st in
      let: st' := l_after_store st l' a i x in
      p (l_pc st) = (Lstore a e x, [:: l'])
      -> eval (dt_vm dt) e = Vint i
      -> in_bounds a i
      -> lstep p st Dstep (Oaddr a i) st'

  | lstep_store_s :
    forall st a e x l' i b j,
      let: dt := l_dt st in
      let: st' := l_after_store st l' b j x in
      p (l_pc st) = (Lstore a e x, [:: l'])
      -> eval (dt_vm dt) e = Vint i
      -> ~~ in_bounds a i
      -> in_bounds b j
      -> dt_ms dt
      -> lstep p st (Dmem b j) (Oaddr a i) st'

  | lstep_init_msf :
    forall st msf l',
      p (l_pc st) =(Linit_msf msf, [::l'])
      -> ~~ dt_ms (l_dt st)
      -> lstep p st Dstep Onone (l_after_init_msf st l' msf)

  | lstep_update_msf :
    forall a st x y e l',
      p (l_pc st) = (Lupdate_msf a x y e, [:: l'])
      -> lstep p st Dstep Onone (l_after_update_msf st l' x y e)

  | lstep_protect :
    forall st x y msf l',
      p (l_pc st) = (Lprotect x y msf, [::l'])
      -> lstep p st Dstep Onone (l_after_protect st l' x y msf)

  | lstep_jump :
    forall a st l',
      p (l_pc st) = (Ljump a, [:: l'])
      -> lstep p st Dstep Onone (with_pc st l')

  | lstep_cjump :
    forall a st e l1 l0 b b',
      p (l_pc st) = (Lcjump a e, [:: l1; l0])
      -> let: dt := l_dt st in
         eval (dt_vm dt) e = Vbool b'
         -> let: l' := if b then l1 else l0 in
            let: st' := with_pc (with_dt st (upd_ms dt (b != b'))) l' in
            lstep p st (Dforce b) (Obranch b') st'.

Inductive lsem
  (p : lprogram) :
  lstate -> seq directive -> seq observation -> lstate -> Prop :=
  | lsem_done :
    forall st, lsem p st [::] [::] st

  | lsem_step :
    forall st st' st'' d o ds os,
      lstep p st d o st'
      -> lsem p st' ds os st''
      -> lsem p st (d :: ds) (o :: os) st''
.

Lemma lstepI (lp:lprogram) (st:lstate) (d:directive) (o:observation) (st':lstate) :
  lstep lp st d o st' ->
  match lp (l_pc st) with
  | (Lassign a x e, [::l']) =>
    [/\ d = Dstep, o = Onone & st' = l_after_assign st l' x e ]

  | (Lload x a e, [::l']) =>
    let: dt := l_dt st in
    exists i b j,
    [/\ eval (dt_vm dt) e = Vint i
      , in_bounds b j
      , (if in_bounds a i
         then [/\ a = b, i = j & d = Dstep ]
         else [/\ dt_ms dt & d = Dmem b j ])
      , o = Oaddr a i
      & st' = l_after_load st l' x b j]

  | (Lstore a e x, [::l']) =>
    let: dt := l_dt st in
    exists i b j,
    [/\ eval (dt_vm dt) e = Vint i
      , in_bounds b j
      , (if in_bounds a i
         then [/\ a = b, i = j & d = Dstep ]
         else [/\ dt_ms dt & d = Dmem b j ])
      , o = Oaddr a i
      & st' = l_after_store st l' b j x ]

  | (Linit_msf msf, [::l']) =>
    [/\ ~~ dt_ms (l_dt st)
      , d = Dstep, o = Onone & st' = l_after_init_msf st l' msf ]

  | (Lupdate_msf a x y e, [:: l']) =>
    [/\ d = Dstep, o = Onone & st' = l_after_update_msf st l' x y e ]

  | (Lprotect x y msf, [::l']) =>
    [/\ d = Dstep, o = Onone & st' = l_after_protect st l' x y msf ]

  | (Ljump a, [:: l']) =>
    [/\ d = Dstep, o = Onone & st' = with_pc st l' ]

  | (Lcjump a e, [:: l1; l0]) =>
    let: dt := l_dt st in
    exists b b',
    let: l' := if b then l1 else l0 in
    [/\ eval (dt_vm dt) e = Vbool b'
      , d = Dforce b, o = Obranch b' & st' = with_pc (with_dt st (upd_ms dt (b != b'))) l' ]

  | _ => False
  end.
Proof.
  case => {st d o st'}; try by move=> > ->.
  + move=> st x a e l' i -> -> hin. exists i, a, i. rewrite hin. by eeauto.
  + move=> st x a e l' i b j -> -> hin ??.
    exists i, b, j.
    rewrite (negbTE hin).
    by eeauto.
  + move=> st a e x l' i -> -> hin. exists i, a, i. rewrite hin. by eeauto.
  + move=> st a e x l' i b j -> -> hin ??.
    exists i, b, j.
    rewrite (negbTE hin).
    by eeauto.
  by move=> a st e l1 l0 b b' -> ->; exists b, b'.
Qed.

Definition l_speculative_constant_time
  (phi : lstate -> lstate -> Prop) (lp : lprogram) : Prop :=
  forall ds os1 os2 st1 st2 st1' st2',
    phi st1 st2 ->
    lsem lp st1 ds os1 st1' ->
    lsem lp st2 ds os2 st2' ->
    os1 = os2.

Definition eq_on_RA (RA : Sr.t) (vm ra_vm : Vm.t) : Prop :=
  eq_map
    (fun x => if Sr.mem x RA then tpublic else tsecret)
    (Vm.get vm)
    (Vm.get ra_vm).

Definition is_initial_lstate
  (Lin : funname -> label) (p : program) (st : lstate) : bool :=
  l_pc st == Lin (p_entry_fn p).

Definition l_initial_phi
  (Lin : funname -> label)
  (RA : Sr.t)
  (p : program)
  (gamma : ctx)
  (st1 st2 : lstate) :
  Prop :=
  let: dt1 := l_dt st1 in
  let: dt2 := l_dt st2 in
  [/\ is_initial_lstate Lin p st1
    , is_initial_lstate Lin p st2
    , initial_phi_dt gamma dt1 dt2
    & eq_on_RA RA (dt_vm dt1) (dt_vm dt2)
  ].
