(* -------------------------------------------------------------------------- *)
(* Security. *)
(* -------------------------------------------------------------------------- *)

From Coq Require Import ZArith.
From mathcomp Require Import all_ssreflect.

Require Import
  semantics
  syntax
  typesystem
  utils
.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Open Scope Z.


(* -------------------------------------------------------------------------- *)
(* Speculative constant time (phi-SCT). *)

Section SCT.

Context (phi : state -> state -> Prop).

Definition speculative_constant_time (p : program) : Prop :=
  forall ds os1 os2 st1 st2 st1' st2',
    phi st1 st2 ->
    sem p st1 ds os1 st1' ->
    sem p st2 ds os2 st2' ->
    os1 = os2.

End SCT.


(* -------------------------------------------------------------------------- *)
(* Type interpretation (indistinguishability). *)

(* Non-secret (public and free) variables need to coincide. *)
Definition eq_val (t : vtype) (x y : value) : bool :=
  [|| ~~ ty_eqb t tpublic | x == y ].

Section EQ_MAP.

Context (is_n : bool).

Definition proj_is_n x := if is_n then sty_n x else to_ty (sty_s x).

Definition eq_map (X : Type) (get : X -> vtype) (st st' : X -> value) : Prop :=
  forall x, eq_val (get x) (st x) (st' x).

Definition eq_vm (gamma : ctx) (vm0 vm1 : Vm.t) : Prop :=
  eq_map (fun x => proj_is_n (ctx_reg gamma x)) (Vm.get vm0) (Vm.get vm1).

Definition eq_mem (gamma : ctx) (st0 st1 : Mem.t) : Prop :=
  eq_map
    (fun '(a, _) => proj_is_n (ctx_arr gamma a))
    (uncurry (Mem.read st0))
    (uncurry (Mem.read st1)).

Record eq_dt_on (gamma : ctx) (dt dt' : data) : Prop :=
  {
    eq_dt_on_vm : eq_vm gamma (dt_vm dt) (dt_vm dt');
    eq_dt_on_mem : eq_mem gamma (dt_mem dt) (dt_mem dt');
  }.

End EQ_MAP.

Definition eq_vm_n := eq_vm true.
Definition eq_vm_s := eq_vm false.
Definition eq_mem_n := eq_mem true.
Definition eq_mem_s := eq_mem false.
Definition eq_dt_on_n := eq_dt_on true.
Definition eq_dt_on_s := eq_dt_on false.

Record eq_dt (gamma : ctx) (dt dt' : data) : Prop :=
  {
    eq_dt_spec : eq_dt_on_s gamma dt dt';
    eq_dt_seq : dt_ms dt \/ eq_dt_on_n gamma dt dt';
    eq_dt_ms : dt_ms dt = dt_ms dt';
  }.

Record eq_st (gamma : ctx) (st st' : state) : Prop :=
  {
    eq_st_c : st_c st = st_c st';
    eq_st_fn : st_fn st = st_fn st';
    eq_st_cs : st_cs st = st_cs st';
    eq_st_dt : eq_dt gamma (st_dt st) (st_dt st');
  }.

Definition initial_phi_vm (gamma : ctx) (vm1 vm2 : Vm.t) : Prop :=
  eq_vm_n gamma vm1 vm2 /\ eq_vm_s gamma vm1 vm2.

Definition initial_phi_mem (gamma : ctx) (mem1 mem2 : Mem.t) : Prop :=
  eq_mem_n gamma mem1 mem2 /\ eq_mem_s gamma mem1 mem2.

Definition initial_phi_dt (gamma : ctx) (dt1 dt2 : data) : Prop :=
  [/\ initial_phi_vm gamma (dt_vm dt1) (dt_vm dt2)
    , initial_phi_mem gamma (dt_mem dt1) (dt_mem dt2)
    & dt_ms dt1 = dt_ms dt2
  ].

Definition initial_phi (p : program) (gamma : ctx) (st1 st2 : state) : Prop :=
  [/\ is_initial_state p st1
    , is_initial_state p st2
    & initial_phi_dt gamma (st_dt st1) (st_dt st2)
  ].

(* Final theorem datament. *)
Definition Typedp_soundness
  (p : program) (cert : certificate) : Prop :=
  let: c := cert (p_entry_fn p) in
  let: sigma := cert_msft c in
  let: gamma := cert_ctx c in
  let: phi := initial_phi p gamma in
  wfcert cert p
  -> typedp cert p
  -> speculative_constant_time phi p.
