From Coq Require Import ZArith.
From Coq Require Morphisms.
From mathcomp Require Import all_ssreflect.

Require Import
  security
  semantics
  semantics_facts
  syntax
  syntax_facts
  typesystem
  typesystem_facts
  utils
  utils_facts
.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Open Scope Z.


(* -------------------------------------------------------------------------- *)
(* Valid states. *)

Section VALID.

Context (cert : certificate).

Notation typedi := (typedi cert).
Notation typedc := (typedc cert).

(* We need to know that the MSF is in range because after an update_msf, if
   we are not misspeculating, we know
       [Vm.get vm1 msf != msf_mask]
       [Vm.get vm2 msf != msf_mask]
   but we need to prove
       [eq_val (Vm.get vm1 msf) (Vm.get vm2 msf)]
   This is true because [eq_vm gamma vm1 vm2] and [gamma msf = Public], but to
   prove this we need to keep the invariant that [msf] is always public.
   Having this in [synced] is less effort (we would need a new invariant on
   [sigma], [gamma] and [vm]. *)
Definition msf_in_range (vm : Vm.t) (x : regvar) : bool :=
  [|| Vm.get vm x == msf_mask | Vm.get vm x == msf_nomask ].

Definition synced (sigma : msft) (vm : Vm.t) (ms : bool) : bool :=
  match sigma with
  | MSFTunknown => true
  | MSFTupdated x => [&& ms == (Vm.get vm x == msf_mask) & msf_in_range vm x ]
  | MSFToutdated x e =>
      [&& if ms
          then [|| eval vm e == Vbool false | Vm.get vm x == msf_mask ]
          else [&& eval vm e == Vbool true & Vm.get vm x != msf_mask ]
        & msf_in_range vm x
      ]
  end.

(* We need this part of the invariant because even thought we know that if we
   return, we go to a continuation (it's a hypothesis for speculative returns
   and provable as an invariant for sequential ones), in sequential execution we
   need not only that there exists an instantiation such that the continuation
   is typable, which is what [typedc_contc] says, but that this instantiation is
   exactly the one we used at the callsite. *)
Inductive typed_callstack : msft -> ctx -> callstack -> Prop :=
  | typed_callstack_nil :
      forall sigma gamma,
        typed_callstack sigma gamma [::]

  | typed_callstack_cons :
      forall f l c cs sigma gamma theta,
        let: sigmaf' := cert_msft' (cert f) in
        let: gammaf' := cert_ctx' (cert f) |> ctx_instantiate theta in
        typedc sigma gamma c sigmaf' gammaf'
        -> typed_callstack sigmaf' gammaf' cs
        -> typed_callstack sigma gamma ((f, l, c) :: cs)
.

Definition typedm (sigma : msft) (gamma : ctx) (st : state) : Prop :=
  exists theta,
    let: c := cert (st_fn st) in
    let: sigmaf' := cert_msft' c in
    let: gammaf' := ctx_instantiate theta (cert_ctx' c) in
    typedc sigma gamma (st_c st) sigmaf' gammaf'
    /\ typed_callstack sigmaf' gammaf' (st_cs st).

End VALID.

Section UTILS.

Context
  (cert : certificate)
  (p : program)
.

Notation typedi := (typedi cert).
Notation typedc := (typedc cert).
Notation step := (step p).
Notation typedm := (typedm cert).
Notation typed_callstack := (typed_callstack cert).

Lemma eq_val_refl t v :
  eq_val t v v.
Proof. by rewrite /eq_val eqxx orbT. Qed.

Lemma weak_eq_val t t' v v' :
  eq_val t v v'
  -> ty_le t t'
  -> eq_val t' v v'.
Proof.
  move=> /orP [].
  - move=> /ty_eqb_tpublic ht hle.
    apply: orb_implP => /ty_eqb_tpublic ?; subst t'.
    move: hle => /ty_le_tpublicI ?.
    by subst t.
  move=> /eqP -> _.
  exact: eq_val_refl.
Qed.

Lemma weak_eq_map X (get get' : X -> vtype) st0 st1 :
  eq_map get st0 st1
  -> (forall x, ty_le (get x) (get' x))
  -> eq_map get' st0 st1.
Proof. move=> ???; eauto using weak_eq_val. Qed.

Lemma weak_eq_vm is_n gamma gamma' vm0 vm1 :
  eq_vm is_n gamma vm0 vm1
  -> ctx_le gamma gamma'
  -> eq_vm is_n gamma' vm0 vm1.
Proof.
  move=> /weak_eq_map h hle.
  apply: h => x.
  case: is_n (ctx_le_reg hle x) => /andP [] //=.
  by eeauto.
Qed.

Lemma weak_eq_mem is_n gamma gamma' st0 st1 :
  eq_mem is_n gamma st0 st1
  -> ctx_le gamma gamma'
  -> eq_mem is_n gamma' st0 st1.
Proof.
  move=> /weak_eq_map h hle.
  apply: h => -[a i].
  case: is_n (ctx_le_arr hle a) => /andP [] //=.
  by eeauto.
Qed.

Lemma weak_eq_dt_on_ms is_n gamma gamma' dt1 dt2 b b' :
  let: dt1' := with_ms dt1 (b || dt_ms dt1) in
  let: dt2' := with_ms dt2 (b' || dt_ms dt2) in
  eq_dt_on is_n gamma dt1 dt2
  -> ctx_le gamma gamma'
  -> eq_dt_on is_n gamma' dt1' dt2'.
Proof.
  move=> [hvm hmem] hle.
  case: b;
    case: b';
    rewrite ?with_ms_id;
    split;
    by eauto using weak_eq_vm, weak_eq_mem.
Qed.

Lemma eq_dt_set_ms gamma dt dt' :
  eq_dt_on_s gamma dt dt'
  -> eq_dt gamma (set_ms dt) (set_ms dt').
Proof. by move=> [hvm hmem]; split=> //; left. Qed.

Lemma eq_dt_with_ms gamma dt1 dt2 b :
  eq_dt gamma dt1 dt2 ->
  eq_dt gamma (with_ms dt1 (b || dt_ms dt1)) (with_ms dt2 (b || dt_ms dt2)).
Proof.
  case b => //=; case: dt1 dt2 => vm1 st1 ms1 [vm2 st2 ms2] h //=.
  by case: h => [/= h1 ?] <-; have /= := eq_dt_set_ms h1; apply.
Qed.

Lemma weak_eq_dt_on is_n gamma gamma' dt dt' :
  eq_dt_on is_n gamma dt dt'
  -> ctx_le gamma gamma'
  -> eq_dt_on is_n gamma' dt dt'.
Proof.
  move=> /(weak_eq_dt_on_ms false false) /[apply]. by rewrite /= !with_ms_id.
Qed.

Lemma weak_eq_dt gamma gamma' dt dt' :
  eq_dt gamma dt dt'
  -> ctx_le gamma gamma'
  -> eq_dt gamma' dt dt'.
Proof.
  move=> [hspec hseq hms] hle.
  split=> //.
  - apply: weak_eq_dt_on; by eauto.
  case: hseq; first by auto.
  move=> /weak_eq_dt_on h.
  right.
  exact: h.
Qed.


Section EQ_CTX_EQ_DT.

Import Morphisms.

#[local]
Lemma eq_ctx_eq_vm :
  Proper (eq ==> eq_ctx ==> eq ==> eq ==> Basics.impl) eq_vm.
Proof. move=> ? _ <- ?? [h _] ? _ <- ? _ <- ??. congruence. Qed.

#[export]
Instance eq_ctx_eq_vm_Proper :
  Proper (eq ==> eq_ctx ==> eq ==> eq ==> iff) eq_vm.
Proof. split; apply: eq_ctx_eq_vm => //; by symmetry. Qed.

#[local]
Lemma eq_ctx_eq_mem :
  Proper (eq ==> eq_ctx ==> eq ==> eq ==> Basics.impl) eq_mem.
Proof.
  move=> ? _ <- ?? [_ hgamma] ? _ <- ? _ <- hmem [??].
  rewrite -hgamma.
  exact: (hmem (_, _)).
Qed.

#[export]
Instance eq_ctx_eq_mem_Proper :
  Proper (eq ==> eq_ctx ==> eq ==> eq ==> iff) eq_mem.
Proof. split; apply: eq_ctx_eq_mem => //; by symmetry. Qed.

#[local]
Lemma eq_ctx_eq_dt_on :
  Proper (eq ==> eq_ctx ==> eq ==> eq ==> Basics.impl) eq_dt_on.
Proof.
  move=> is_n _ <- ?? hgamma ? _ <- ? _ <- [hvm hmem].
  split; by rewrite -hgamma.
Qed.

#[export]
Instance eq_ctx_eq_dt_on_Proper :
  Proper (eq ==> eq_ctx ==> eq ==> eq ==> iff) eq_dt_on.
Proof. split; apply: eq_ctx_eq_dt_on => //; by symmetry. Qed.

#[local]
Lemma eq_ctx_eq_dt :
  Proper (eq_ctx ==> eq ==> eq ==> Basics.impl) eq_dt.
Proof.
  move=> ?? hgamma ? _ <- ? _ <- [hspec hseq hms].
  split=> //.
  - exact: (eq_ctx_eq_dt_on erefl hgamma erefl erefl).
  case: hseq => hseq; first by left.
  right.
  exact: (eq_ctx_eq_dt_on erefl hgamma erefl erefl).
Qed.

#[export]
Instance eq_ctx_eq_dt_Proper :
  Proper (eq_ctx ==> eq ==> eq ==> iff) eq_dt.
Proof. split; apply: eq_ctx_eq_dt => //; by symmetry. Qed.

End EQ_CTX_EQ_DT.

Lemma eq_vm_eval_e_public is_n gamma vm vm' e t :
  eq_vm is_n gamma vm vm'
  -> typede gamma e t
  -> ty_eqb (proj_is_n is_n t) tpublic
  -> eval vm e = eval vm' e.
Proof.
  move=> hvm hty ht.
  apply: eval_eq_on => x /(typede_free_variables hty) {hty}.
  move=> /andP [].
  case: is_n ht hvm (hvm x) => /ty_eqb_tpublic /=.
  - move=> -> _ hx /ty_le_tpublicI h _. move: hx. by rewrite h => /eqP.
  move=> /to_ty_tpublic -> _ hx _ /lvl_le_PublicI h.
  move: hx.
  by rewrite h => /eqP.
Qed.

Lemma eq_vm_vm_set is_n gamma vm vm' x v v' t :
  let: gamma' := ctx_update_reg gamma x t in
  eq_vm is_n gamma vm vm'
  -> eq_val (proj_is_n is_n t) v v'
  -> eq_vm is_n gamma' (Vm.set vm x v) (Vm.set vm' x v').
Proof.
  move=> h hv y.
  rewrite /= !Vm.get_set.
  by case: eqP.
Qed.

Lemma eq_mem_mem_write is_n gamma gamma' st st' a i v v' t :
  let: proj := proj_is_n is_n in
  eq_mem is_n gamma st st'
  -> ctx_le gamma gamma'
  -> ty_le (proj t) (ctx_arr gamma' a |> proj)
  -> in_bounds a i
  -> eq_val (proj t) v v'
  -> eq_mem is_n gamma' (Mem.write st a i v) (Mem.write st' a i v').
Proof.
  move=> h hle ht hi hv [b j].
  rewrite /= !Mem.read_write // {hi}.
  case: andP => [[/eqP <- _] | _]; first by eauto using weak_eq_val.
  apply: weak_eq_val; first exact: (h (_, _)).
  move: (ctx_le_arr hle b) => /andP [].
  case: is_n h ht hv => //.
  by eeauto.
Qed.

Lemma eq_vm_vm_set_eq is_n gamma vm vm' x v :
  eq_vm is_n gamma vm vm'
  -> eq_vm is_n gamma (Vm.set vm x v) (Vm.set vm' x v).
Proof.
  move=> h.
  rewrite -(ctx_update_reg_id gamma x).
  apply: eq_vm_vm_set => //; by rewrite eq_val_refl ?orbT.
Qed.

Lemma eq_dt_on_write_reg is_n gamma dt1 dt2 x v v' t :
  let: gamma' := ctx_update_reg gamma x t in
  eq_dt_on is_n gamma dt1 dt2
  -> eq_val (proj_is_n is_n t) v v'
  -> eq_dt_on is_n gamma' (write_reg dt1 x v) (write_reg dt2 x v').
Proof.
  move=> [hvm hmem].
  split=> //.
  exact: eq_vm_vm_set.
Qed.

Lemma eq_dt_on_write_mem is_n gamma gamma' dt1 dt2 a i v v' t :
  let: proj := proj_is_n is_n in
  eq_dt_on is_n gamma dt1 dt2
  -> ctx_le gamma gamma'
  -> ty_le (proj t) (ctx_arr gamma' a |> proj)
  -> in_bounds a i
  -> eq_val (proj t) v v'
  -> eq_dt_on is_n gamma' (write_mem dt1 a i v) (write_mem dt2 a i v').
Proof.
  move=> [hvm hmem] hgamma hi hv.
  split; first by eauto using weak_eq_vm.
  apply: eq_mem_mem_write; by eauto.
Qed.

Lemma eq_dt_write_reg gamma dt1 dt2 x v v' t :
  let: gamma' := ctx_update_reg gamma x t in
  eq_dt gamma dt1 dt2
  -> eq_val (to_ty (sty_s t)) v v'
  -> (dt_ms dt1 || eq_val (sty_n t) v v')
  -> eq_dt gamma' (write_reg dt1 x v) (write_reg dt2 x v').
Proof.
  move=> [hspec hseq hms] hspecv hseqv.
  split=> //; first exact: eq_dt_on_write_reg.
  case: hseq => hseq; first by left.
  move: hseqv => /orP [] hseqv; first by left.
  right.
  exact: eq_dt_on_write_reg.
Qed.

Lemma eq_dt_write_mem gamma gamma' dt1 dt2 a i v v' t :
  eq_dt gamma dt1 dt2
  -> ctx_le gamma gamma'
  -> sty_le t (ctx_arr gamma' a)
  -> in_bounds a i
  -> eq_val (to_ty (sty_s t)) v v'
  -> (dt_ms dt1 || eq_val (sty_n t) v v')
  -> eq_dt gamma' (write_mem dt1 a i v) (write_mem dt2 a i v').
Proof.
  move=> [hspec hseq hms] hle /andP [htn hts] hi hspecv hseqv.
  split=> //.
  - apply: eq_dt_on_write_mem; by eauto using ty_le_to_ty.
  case: hseq => hseq; first by left.
  move: hseqv => /orP [] hseqt; first by left.
  right; apply: eq_dt_on_write_mem; by eauto.
Qed.

Lemma eq_dt_write_reg_eq gamma dt1 dt2 x v :
  eq_dt gamma dt1 dt2
  -> eq_dt gamma (write_reg dt1 x v) (write_reg dt2 x v).
Proof.
  move=> h.
  rewrite -(ctx_update_reg_id gamma x).
  apply: eq_dt_write_reg => //; by rewrite eq_val_refl ?orbT.
Qed.

Lemma eq_dt_on_ctx_update_write_reg is_n gamma dt1 dt2 x e t :
  let: gamma' := ctx_update_reg gamma x t in
  let: dt1' := write_reg dt1 x (eval (dt_vm dt1) e) in
  let: dt2' := write_reg dt2 x (eval (dt_vm dt2) e) in
  eq_dt_on is_n gamma dt1 dt2
  -> typede gamma e t
  -> eq_dt_on is_n gamma' dt1' dt2'.
Proof.
  move=> [hvm hmem] he.
  apply: eq_dt_on_write_reg => //.
  apply: orb_implP => ht.
  apply/eqP.
  apply: eq_vm_eval_e_public; by eauto.
Qed.

Lemma eq_dt_ctx_update_write_reg gamma dt1 dt2 x e t :
  let: gamma' := ctx_update_reg gamma x t in
  let: dt1' := write_reg dt1 x (eval (dt_vm dt1) e) in
  let: dt2' := write_reg dt2 x (eval (dt_vm dt2) e) in
  eq_dt gamma dt1 dt2
  -> typede gamma e t
  -> eq_dt gamma' dt1' dt2'.
Proof.
  move=> [hspec hseq hms] he.
  split=> //; first exact: eq_dt_on_ctx_update_write_reg.
  case: hseq; first by left.
  right.
  exact: eq_dt_on_ctx_update_write_reg.
Qed.

Lemma eq_mem_ctx_update_reg is_n gamma x t st st' :
  eq_mem is_n gamma st st'
  -> eq_mem is_n (ctx_update_reg gamma x t) st st'.
Proof. done. Qed.

Lemma eq_vm_s_instantiate vm vm' theta theta' gamma gamma' :
  eq_vm_s gamma vm vm'
  -> ctx_le gamma (ctx_instantiate theta gamma')
  -> eq_vm_s (ctx_instantiate theta' gamma') vm vm'.
Proof.
  move=> hvm hle.
  apply: (weak_eq_map hvm) => x.
  move: (ctx_le_reg hle x) => /andP [] _.
  by rewrite ty_le_to_ty.
Qed.

Lemma eq_mem_s_instantiate st st' theta theta' gamma gamma' :
  eq_mem_s gamma st st'
  -> ctx_le gamma (ctx_instantiate theta gamma')
  -> eq_mem_s (ctx_instantiate theta' gamma') st st'.
Proof.
  move=> hvm hle.
  apply: (weak_eq_map hvm) => -[a i].
  move: (ctx_le_arr hle a) => /andP [] _.
  by rewrite ty_le_to_ty.
Qed.

Lemma eq_dt_on_s_instantiate dt1 dt2 gamma gamma' xi theta theta' :
  eq_dt_on_s gamma dt1 dt2
  -> ctx_le gamma (ctx_instantiate theta gamma')
  -> eq_dt_on_s
       (ctx_instantiate theta' gamma')
       (dt_after_RET_S dt1 xi)
       (dt_after_RET_S dt2 xi).
Proof.
  move=> [hvm hmem] hle.
  split.
  - case: xi => [msf|] /=.
    + apply: eq_vm_vm_set_eq. apply: eq_vm_s_instantiate; by eauto.
    apply: eq_vm_s_instantiate; by eauto.
  apply: eq_mem_s_instantiate; by eauto.
Qed.

Lemma weak_synced sigma sigma' vm ms :
  synced sigma vm ms
  -> msft_le sigma' sigma
  -> synced sigma' vm ms.
Proof. move=> h [] ?; by subst. Qed.

Lemma synced_msft_cond sigma vm e b b' ms :
   synced sigma vm ms
  -> eval vm e = Vbool b'
  -> synced (msft_cond sigma (if b then e else enot e)) vm ((b != b') || ms).
Proof.
  case: sigma => [//| msf |//] /= /andP [/eqP <- ->].
  by case: b b' => /= -[] /= -> //=; case: ms.
Qed.

Lemma synced_msft_restrict sigma vm x v ms :
  synced sigma vm ms
  -> synced (msft_restrict sigma x) (Vm.set vm x v) ms.
Proof.
  rewrite /msft_restrict.
  case hx: Sr.mem => //.
  move: hx => /negPf.
  case: sigma => [// | msf | msf e] /=.
  - by rewrite /msf_in_range Sr.F.singleton_b Sr.eqbP eq_sym Vm.get_set => ->.
  rewrite Sr.F.add_b Sr.eqbP eq_sym => /norP [??].
  by rewrite /msf_in_range eval_set // set_neq.
Qed.

Lemma typed_callstackI sigma gamma cs :
  typed_callstack sigma gamma cs
  -> match cs with
     | [::] => True
     | (f, _, c) :: cs' =>
         exists theta,
           let: sigmaf' := cert_msft' (cert f) in
           let: gammaf' := cert_ctx' (cert f) |> ctx_instantiate theta in
           typedc sigma gamma c sigmaf' gammaf'
           /\ typed_callstack sigmaf' gammaf' cs'
     end.
Proof. case; by eeauto. Qed.

Lemma is_n_sty_after_init_msf is_n t :
  let: t' := proj_is_n is_n (sty_after_init_msf t) in
  t' = sty_n t \/ t' = tsecret.
Proof. case: is_n => /=; first by left. case: sty_n => [|[]]; by auto. Qed.

Lemma eq_dt_on_ctx_after_init_msf is_n gamma dt1 dt2 msf :
  let: gamma' := ctx_after_init_msf gamma msf in
  let: dt1' := write_reg dt1 msf msf_nomask in
  let: dt2' := write_reg dt2 msf msf_nomask in
  eq_dt_on_n gamma dt1 dt2
  -> eq_dt_on is_n gamma' dt1' dt2'.
Proof.
  move=> [hvm hmem].
  split=> [x | [a i]] /=.
  - rewrite !Vm.get_set.
    case: eqP => _; first exact: eq_val_refl.
    by case: (is_n_sty_after_init_msf is_n (ctx_reg gamma x)) => ->.
  case: (is_n_sty_after_init_msf is_n (ctx_arr gamma a)) => -> //.
  exact: (hmem (_, _)).
Qed.

Lemma eq_dt_ctx_after_init_msf gamma dt1 dt2 msf :
  let: gamma' := ctx_after_init_msf gamma msf in
  let: dt1' := write_reg dt1 msf msf_nomask in
  let: dt2' := write_reg dt2 msf msf_nomask in
  eq_dt gamma dt1 dt2
  -> ~~ dt_ms dt1
  -> eq_dt gamma' dt1' dt2'.
Proof.
  move=> [_ hseq hms] /negPf hms1.
  case: hseq => hseq; first congruence.
  split=> //; first exact: eq_dt_on_ctx_after_init_msf.
  right.
  exact: eq_dt_on_ctx_after_init_msf.
Qed.

Lemma eq_dt_ctx_after_update_msf gamma dt1 dt2 x y e :
  let: dt1' := dt_after_update_msf dt1 x y e in
  let: dt2' := dt_after_update_msf dt2 x y e in
  eq_dt gamma dt1 dt2
  -> synced (MSFToutdated y e) (dt_vm dt1) (dt_ms dt1)
  -> synced (MSFToutdated y e) (dt_vm dt2) (dt_ms dt2)
  -> eq_dt gamma dt1' dt2'.
Proof.
  move=> hst.
  have := eq_dt_ms hst.
  rewrite /dt_after_update_msf /=.
  case: (dt_ms dt1) => /= <-.
  - move=>  /andP [] /orP [] /eqP -> _ /andP [] /orP [] /eqP -> _ /=;
      rewrite ?if_same;
      exact: eq_dt_write_reg_eq.
  rewrite /msf_in_range => /andP [] /andP [] /eqP -> /negPf -> /eqP -> /andP []
    /andP [] /eqP -> /negPf -> /eqP ->.
  exact: eq_dt_write_reg_eq.
Qed.

Lemma eq_dt_ctx_after_protect gamma dt1 dt2 x y msf :
  let: dt1' := dt_after_protect dt1 x y msf in
  let: dt2' := dt_after_protect dt2 x y msf in
  eq_dt gamma dt1 dt2
  -> synced (MSFTupdated msf) (dt_vm dt1) (dt_ms dt1)
  -> synced (MSFTupdated msf) (dt_vm dt2) (dt_ms dt2)
  -> eq_dt (ctx_after_protect gamma x y) dt1' dt2'.
Proof.
  rewrite /dt_after_protect.
  move=> [hspec hseq hms] /andP [/eqP <- _] /andP [/eqP <- _].
  rewrite -hms.
  case hms1: dt_ms.
  - apply: eq_dt_write_reg => //; first exact: eq_val_refl. by rewrite hms1.
  move: hseq.
  rewrite hms1.
  move=> [// | hseq].
  apply: eq_dt_write_reg.
  - split=> //. by right.
  - rewrite /sty_after_init_msf /=.
    apply: weak_eq_val; first exact: (eq_dt_on_vm hseq).
    exact: ty_le_to_lvl_to_ty.
  rewrite hms1 /=.
  exact: (eq_dt_on_vm hseq).
Qed.

Lemma synced_after_ret_s sigma sigma' xi dt :
  msft_call sigma xi = Some sigma'
  -> synced sigma' (dt_vm (dt_after_RET_S dt xi)) true.
Proof.
  move=> /msft_callE [[-> ->] // | [msf [-> -> ->]]]; subst.
  by rewrite /= /msf_in_range set_eq.
Qed.

Lemma synced_after_init_msf vm msf :
  synced (MSFTupdated msf) (Vm.set vm msf msf_nomask) false.
Proof. by rewrite /= /msf_in_range set_eq. Qed.

Lemma synced_after_update_msf dt x y e :
  let: dt' := dt_after_update_msf dt x y e in
  synced (MSFToutdated y e) (dt_vm dt) (dt_ms dt)
  -> synced (MSFTupdated x) (dt_vm dt') (dt_ms dt).
Proof.
  move=> /andP [].
  rewrite /= /msf_in_range set_eq.
  case: dt_ms; last by move=> /andP [/eqP -> /negPf ->].
  move=> /orP [] /eqP ->; first by rewrite eqxx.
  by case: ifP.
Qed.

End UTILS.


Section TYPEDM_SYNCED.

Context
  (cert : certificate)
  (p : program)
  (htyp : typedp cert p)
.

Lemma typedm_synced_step st st' d o sigma gamma :
  let: dt := st_dt st in
  let: dt' := st_dt st' in
  step p st d o st' ->
  typedm cert sigma gamma st ->
  synced sigma (dt_vm dt) (dt_ms dt) ->
  exists sigma' gamma',
      [/\ typedm cert sigma' gamma' st'
        & synced sigma' (dt_vm dt') (dt_ms dt')
      ].
Proof.
  case=> {st st' d o} => - [_c g cs dt] /=.

  - move=> x e c ? [theta [/= htyc htycs]] hsync; subst _c.
    move: htyc => /typedcI [sigma0 [gamma0 [? [? [? [? [htyi htyc ????]]]]]]].
    move: htyi => /typediI [? [t htye ?]]; subst.
    exists (msft_restrict sigma0 x), (ctx_update_reg gamma0 x t); split.
    + exists theta; split=> /=; last assumption. apply: typed_weak; by eeauto.
    apply: synced_msft_restrict. apply: weak_synced; eassumption.

  - move=> x a e c i ? hi ha [theta [/= htyc htycs]] hsync; subst _c.
    move: htyc => /typedcI [sigma0 [gamma0 [? [? [? [? [htyi htyc ????]]]]]]].
    move: htyi => /typediI [htye ??]; subst.
    exists (msft_restrict sigma0 x), (ctx_after_load gamma0 x a); split.
    + exists theta; split=> /=; last assumption. apply: typed_weak; by eeauto.
    apply: synced_msft_restrict. apply: weak_synced; eassumption.

  - move=> x a e c i b j ? hi hb ha hms [theta [/= htyc htycs]] hsync; subst _c.
    move: htyc => /typedcI [sigma0 [gamma0 [? [? [? [? [htyi htyc ????]]]]]]].
    move: htyi => /typediI [htye ??]; subst.
    exists (msft_restrict sigma0 x), (ctx_after_load gamma0 x a); split.
    + exists theta; split=> /=; last assumption. apply: typed_weak; by eeauto.
    apply: synced_msft_restrict. apply: weak_synced; eassumption.

  - move=> a e x c i ? hi ha [theta [/= htyc htycs]] hsync; subst _c.
    move: htyc => /typedcI [sigma0 [? [? [gamma1 [? [? [htyi htyc ????]]]]]]].
    move: htyi => /typediI [htye ? hgamma1]; subst.
    exists sigma0, gamma1; split.
    + exists theta; split=> /=; last assumption. apply: typed_weak; by eeauto.
    apply: weak_synced; eassumption.

  - move=> a e x c i b j ? hi hb ha hms [theta [/= htyc htycs]] hsync; subst _c.
    move: htyc => /typedcI [sigma0 [? [? [gamma1 [? [? [htyi htyc ????]]]]]]].
    move: htyi => /typediI [htye ? hgamma1]; subst.
    exists sigma0, gamma1; split.
    + exists theta; split=> /=; last assumption. apply: typed_weak; by eeauto.
    apply: weak_synced; eassumption.

  - move=> msf c ? /negPf hms [theta [/= htyc htycs]] hsync; subst _c.
    move: htyc => /typedcI [sigma0 [gamma0 [? [? [? [? [htyi htyc ????]]]]]]].
    move: htyi => /typediI [??]; subst.
    exists (MSFTupdated msf), (ctx_after_init_msf gamma0 msf); split.
    + exists theta; split=> /=; last assumption. apply: typed_weak; by eeauto.
    rewrite hms. exact: synced_after_init_msf.

  - move=> msf' msf e c ? [theta [/= htyc htycs]] hsync; subst _c.
    move: htyc => /typedcI [sigma0 [gamma0 [? [? [? [? [htyi htyc ????]]]]]]].
    move: htyi => /typediI [???]; subst.
    exists (MSFTupdated msf'), gamma0; split.
    + exists theta; split=> /=; last assumption. apply: typed_weak; by eeauto.
    apply: synced_after_update_msf. apply: weak_synced; eassumption.

  - move=> x y msf c ? [theta [/= htyc htycs]] hsync; subst _c.
    move: htyc => /typedcI [sigma0 [gamma0 [? [? [? [? [htyi htyc ????]]]]]]].
    move: htyi => /typediI [???]; subst.
    exists (msft_restrict (MSFTupdated msf) x), (ctx_after_protect gamma0 x y);
      split.
    + exists theta; split=> /=; last assumption. apply: typed_weak; by eeauto.
    apply: synced_msft_restrict. apply: weak_synced; eassumption.

  - move=> e c1 c0 c b b' ? he [theta [/= htyc htycs]] hsync; subst _c.
    move: htyc => /typedcI [sigma0 [gamma0 [? [? [? [? [htyi htyc ????]]]]]]].
    move: htyi => /typediI [htye hty1 hty0]; subst.
    exists (msft_cond sigma0 (if b then e else enot e)), gamma0; split.
    + exists theta; split=> /=; last assumption.
      apply: typed_after_if; [eassumption | eassumption|].
      apply: typed_weak; by eeauto.
    apply: synced_msft_cond => //. apply: weak_synced; eassumption.

  - move=> e c1 c b b' ? he [theta [/= htyc htycs]] hsync; subst _c.
    move: htyc => /typedcI [sigma0 [gamma0 [? [? [? [? [htyi htyc ????]]]]]]].
    move: htyi => /typediI [htye hty1 ??]; subst.
    exists (msft_cond sigma0 (if b then e else enot e)), gamma0; split.
    + rewrite -cat_rcons -cats1 -{3}(cat0s c) -(fun_if (fun x => x ++ c)).
      exists theta; split=> /=; last assumption.
      apply: typed_after_if; [| by constructor |].
      * apply: typed_capp; first eassumption.
        apply: typed_weak; by eeauto.
      apply: typed_weak; by eeauto.
    apply: synced_msft_cond => //. apply: weak_synced; eassumption.

  - move=> f l xi c ? [theta [/= htyc htycs]] hsync; subst _c.
    move: htyc => /typedcI [sigma0 [gamma0 [? [? [? [? [htyi htyc ????]]]]]]].
    move: htyi => /typediI [?? [thetaf [??]]]; subst.
    exists (cert_msft (cert f)), (ctx_instantiate thetaf (cert_ctx (cert f)));
      split.
    + exists thetaf; split=> /=; first exact: ctx_instantiate_typedc.
      econstructor; last eassumption.
      apply: typed_weak; by eeauto.
    apply: weak_synced; eassumption.

  move=> f l xi cont st' ? hret [theta [/= htyc htycs]] hsync; subst _c.
  move: htyc => /typedcI [??].
  case: (boolP (is_speculative_ret _ _ _ _)) hret => hret.
  - move=> [/in_contpI hcont ?]; subst st' => /=.
    have [theta' [sigma_call [hsigma_call htycont]]] :=
      typedc_contc (htyp f) hcont.
    exists sigma_call, (ctx_instantiate theta' (cert_ctx' (cert g))); split.
    + exists theta_Secret; split=> /=; last by constructor.
      apply: typed_weak; last exact: ctx_instantiate_theta_Secret; by eeauto.
    apply: synced_after_ret_s. eassumption.

  move=> ?; subst st' => /=.
  have hcs : cs = (f, l, cont) :: drop 1 cs.
  - case: (cs) hret => //= -[[f' l'] cont'] ?.
    by rewrite /is_speculative_ret /= negbK drop0 => /eqP ->.
  move: htycs.
  rewrite hcs => /typed_callstackI [thetaf [htyc {}hcs]].
  exists (cert_msft' (cert g)), (ctx_instantiate theta (cert_ctx' (cert g)));
    split.
  - rewrite /= drop0. by exists thetaf.
  apply: weak_synced; eassumption.
Qed.

Lemma typedm_synced_sem st st' ds os sigma gamma :
  let: dt := st_dt st in
  let: dt' := st_dt st' in
  sem p st ds os st' ->
  typedm cert sigma gamma st ->
  synced sigma (dt_vm dt) (dt_ms dt) ->
  exists sigma' gamma',
    [/\ typedm cert sigma' gamma' st'
      & synced sigma' (dt_vm dt') (dt_ms dt')
    ].
Proof.
  elim: ds st st' os sigma gamma => [|d ds hind] st st' os sigma gamma /semI [].
  - move=> ->. by eeauto.
  move=> o [os' [? [mi hstep hsem]]] htym hsync.
  have [? [? [??]]] := typedm_synced_step hstep htym hsync.
  by eeauto.
Qed.

End TYPEDM_SYNCED.


Module __SingleStepSoundness.

Section PROOF.

Context
  (cert : certificate)
  (p : program)
  (htyp : typedp cert p)
.

Notation typedi := (typedi cert).
Notation typedc := (typedc cert).
Notation step := (step p).
Notation sem := (sem p).
Notation typedm := (typedm cert).

Let sss_invariant sigma gamma st1 st2 :=
  let: dt1 := st_dt st1 in
  let: dt2 := st_dt st2 in
  [/\ typedm sigma gamma st1
    , synced sigma (dt_vm dt1) (dt_ms dt1)
    , synced sigma (dt_vm dt2) (dt_ms dt2)
    & eq_st gamma st1 st2
  ].

Let single_step_soundnessT st1 d :=
  forall st1' st2 st2' o1 o2 sigma gamma,
    step st1 d o1 st1'
    -> step st2 d o2 st2'
    -> sss_invariant sigma gamma st1 st2
    -> o1 = o2
       /\ exists sigma' gamma', sss_invariant sigma' gamma' st1' st2'.

Lemma sss_assign st1 d x e c :
  st_c st1 = Iassign x e :: c
  -> d = Dstep
  -> single_step_soundnessT st1 d.
Proof.
  move: st1 => [/= _ g cs [vm1 st1 ms]] -> ?; subst.
  move=> st1' [/= c2 g2 cs2 [vm2 st2 ms2]] st2' o1 o2 sigma gamma hsem1 hsem2
    [/= hty1 hsync1 hsync2 [/= ??? hst]];
    subst.
  move: hsem1 hsem2 => /stepI [??] /stepI [??]; subst.
  split=> //.
  move: hty1 => [/= theta [hty1 hcs]].
  move: hty1 =>
    /typedcI [sigma0 [gamma0 [sigmai [gammai [sigma0' [gamma0' []]]]]]]
    /typediI [? [t htye ?]] htyc ????;
    subst.
  exists (msft_restrict sigma0 x), (ctx_update_reg gamma0 x t).
  split=> /=; last split=> //=.
  - by eeauto.
  - by eauto using synced_msft_restrict, weak_synced.
  - by eauto using synced_msft_restrict, weak_synced.
  by eauto using eq_dt_ctx_update_write_reg, weak_eq_dt.
Qed.

Lemma sss_load_n st1 d x a e c :
  st_c st1 = Iload x a e :: c
  -> d = Dstep
  -> single_step_soundnessT st1 d.
Proof.
  move: st1 => [/= _ g cs [vm1 st1 ms]] -> ?; subst.
  move=> st1' [/= c2 g2 cs2 [vm2 st2 ms2]] st2' o1 o2 sigma gamma hsem1 hsem2
    [/= hty1 hsync1 hsync2 [/= ??? [/= hspec hseq ?]]];
    subst c2 g2 cs2 ms2.
  move: hsem1 hsem2 =>
    /stepI [i1 [/= heval1 hina1 ??]] /stepI [i2 [/= heval2 _ ??]]; subst.
  move: hty1 => [/= theta [hty1 hcs]].
  move: hty1 =>
    /typedcI [sigma0 [gamma0 [sigmai [gammai [sigma0' [gamma0' []]]]]]]
    /typediI [htye ??] htyc ?? hgamma ?;
    subst.
  move: heval1.
  rewrite (eq_vm_eval_e_public (is_n := false) (vm' := vm2) _ htye erefl);
    first last.
  - move: hspec => [/= ? _]. by eauto using weak_eq_vm.
  rewrite heval2 => -[?]; subst i2.
  split=> //.
  exists (msft_restrict sigma0 x), (ctx_after_load gamma0 x a).
  split=> /=; last split=> //=.
  - by eeauto.
  - by eauto using synced_msft_restrict, weak_synced.
  - by eauto using synced_msft_restrict, weak_synced.
  apply: eq_dt_write_reg; [apply: weak_eq_dt; by eeauto | done|].
  apply/orP.
  case: hseq => [-> | hseq]; first by left.
  right.
  apply: weak_eq_val; first exact: (eq_dt_on_mem hseq (_, _)).
  by move: (ctx_le_arr hgamma a) => /andP [-> _].
Qed.

Lemma sss_load_s st1 d x a e c b j :
  st_c st1 = Iload x a e :: c
  -> d = Dmem b j
  -> single_step_soundnessT st1 d.
Proof.
  move: st1 => [/= _ g cs [vm1 st1 ms]] -> ?; subst.
  move=> st1' [/= c2 g2 cs2 [vm2 st2 ms2]] st2' o1 o2 sigma gamma hsem1 hsem2
    [/= hty1 hsync1 hsync2 [/= ??? [/= hspec hseq ?]]];
    subst c2 g2 cs2 ms2.
  move: hsem1 hsem2 =>
    /stepI [i1 [/= heval1 _ hinb1 hspec1 ??]]
    /stepI [i2 [/= heval2 _ _ _ ??]]; subst.
  move: hty1 => [/= theta [hty1 hcs]].
  move: hty1 =>
    /typedcI [sigma0 [gamma0 [sigmai [gammai [sigma0' [gamma0' []]]]]]]
    /typediI [htye ??] htyc ?? hgamma ?;
    subst.
  move: heval1.
  rewrite (eq_vm_eval_e_public (is_n := false) (vm' := vm2) _ htye erefl);
    first last.
  - move: hspec => [/= ? _]. by eauto using weak_eq_vm.
  rewrite heval2 => -[?]; subst i2.
  split=> //.
  exists (msft_restrict sigma0 x), (ctx_after_load gamma0 x a).
  split=> /=; last split=> //=.
  - by eeauto.
  - by eauto using synced_msft_restrict, weak_synced.
  - by eauto using synced_msft_restrict, weak_synced.
  apply: eq_dt_write_reg; [apply: weak_eq_dt; by eeauto | done|].
  by rewrite /= hspec1.
Qed.

Lemma sss_store_n st1 d x a e c :
  st_c st1 = Istore a e x :: c
  -> d = Dstep
  -> single_step_soundnessT st1 d.
Proof.
  move: st1 => [/= _ g cs [vm1 st1 ms]] -> ?; subst.
  move=> st1' [/= c2 g2 cs2 [vm2 st2 ms2]] st2' o1 o2 sigma gamma hsem1 hsem2
    [/= hty1 hsync1 hsync2 [/= ??? [/= hspec hseq ?]]];
    subst c2 g2 cs2 ms2.
  move: hsem1 hsem2 =>
    /stepI [i1 [/= heval1 hina ??]] /stepI [i2 [/= heval2 _ ??]]; subst.
  move: hty1 => [/= theta [hty1 hcs]].
  move: hty1 =>
    /typedcI [sigma0 [gamma0 [sigmai [gammai [sigma0' [gamma0' []]]]]]]
    /typediI [htye ? [hle /andP [hxn hxs] hb]] htyc ?? hgamma hgamma';
    subst.
  move: heval1.
  rewrite (eq_vm_eval_e_public (is_n := false) (vm' := vm2) _ htye erefl);
    first last.
  - move: hspec => [/= ? _]. by eauto using weak_eq_vm.
  rewrite heval2 => -[?]; subst i2.
  split=> //.
  exists sigma0, gammai.
  split=> /=; last split=> //=.
  - by eeauto.
  - by eauto using weak_synced.
  - by eauto using weak_synced.
  split=> //=.
  + apply: eq_dt_on_write_mem; [| eassumption | | done |].
    + by apply: (weak_eq_dt_on_ms false false hspec).
    + by eeauto.
    apply: weak_eq_val; first exact: (eq_dt_on_vm hspec).
    move: (ctx_le_reg hgamma x) => /andP [??].
    by rewrite ty_le_to_ty; eeauto.
  case: hseq => [? | hseq]; [by left | right].
  apply: eq_dt_on_write_mem; [| eassumption | | done |].
  + by apply: (weak_eq_dt_on_ms false false hseq).
  + by eeauto.
  apply: weak_eq_val; first exact: (eq_dt_on_vm hseq).
  by move: (ctx_le_reg hgamma x) => /andP [].
Qed.

Lemma sss_store_s st1 d x a e c b j :
  st_c st1 = Istore a e x :: c
  -> d = Dmem b j
  -> single_step_soundnessT st1 d.
Proof.
  move: st1 => [/= _ g cs [vm1 st1 ms]] -> ?; subst.
  move=> st1' [/= c2 g2 cs2 [vm2 st2 ms2]] st2' o1 o2 sigma gamma hsem1 hsem2
    [/= hty1 hsync1 hsync2 [/= ??? [/= hspec hseq ?]]];
    subst c2 g2 cs2 ms2.
  move: hsem1 hsem2 =>
    /stepI [i1 [/= heval1 _ hinb hspec1 ??]]
    /stepI [i2 [/= heval2 _ _ _ ??]]; subst.
  move: hty1 => [/= theta [hty1 hcs]].
  move: hty1 =>
    /typedcI [sigma0 [gamma0 [sigmai [gammai [sigma0' [gamma0' []]]]]]]
    /typediI [htye ? [hle /andP [hxn hxs] hb]] htyc ?? hgamma hgamma';
    subst.
  move: heval1.
  rewrite (eq_vm_eval_e_public (is_n := false) (vm' := vm2) _ htye erefl);
    first last.
  - move: hspec => [/= ? _]. by eauto using weak_eq_vm.
  rewrite heval2 => -[?]; subst i2.
  split=> //.
  exists sigma0, gammai.
  split=> /=; last split=> //=.
  - by eeauto.
  - by eauto using weak_synced.
  - by eauto using weak_synced.
  split => //; rewrite hspec1; last by left.
  apply: eq_dt_on_write_mem; [| eassumption | | done |].
  - by apply: (weak_eq_dt_on_ms true true hspec).
  - by eeauto.
  apply: weak_eq_val; first exact: (eq_dt_on_vm hspec).
  move: (ctx_le_reg hgamma x) => /andP [??] /=.
  rewrite ty_le_to_ty.
  case: (a =P b) => [? | /eqP].
  + subst b; by apply: lvl_le_trans hxs.
  by move=> /hb; apply: lvl_le_trans.
Qed.

Lemma sss_init_msf st1 d msf c :
  st_c st1 = Iinit_msf msf :: c
  -> d = Dstep
  -> single_step_soundnessT st1 d.
Proof.
  move: st1 => [/= _ g cs [vm1 st1 ms]] -> ?; subst.
  move=> st1' [/= c2 g2 cs2 [vm2 st2 ms2]] st2' o1 o2 sigma gamma hsem1 hsem2
    [/= hty1 hsync1 hsync2 [/= ??? hst]];
    subst.
  move: hsem1 hsem2 => /stepI [/= /negPf ???] /stepI [/= /negPf ???]; subst.
  split=> //.
  move: hty1 => [/= theta [hty1 hcs]].
  move: hty1 =>
    /typedcI [sigma0 [gamma0 [sigmai [gammai [sigma0' [gamma0' []]]]]]]
    /typediI [??] htyc ????;
    subst.
  exists (MSFTupdated msf), (ctx_after_init_msf gamma0 msf).
  split; last split=> //=.
  - by eeauto.
  - exact: synced_after_init_msf.
  - exact: synced_after_init_msf.
  by eauto using eq_dt_ctx_after_init_msf, weak_eq_dt.
Qed.

Lemma sss_update_msf st1 d x y e c :
  st_c st1 = Iupdate_msf x y e :: c
  -> d = Dstep
  -> single_step_soundnessT st1 d.
Proof.
  move: st1 => [/= _ g cs [vm1 st1 ms]] -> ?; subst.
  move=> st1' [/= c2 g2 cs2 [vm2 st2 ms2]] st2' o1 o2 sigma gamma hsem1 hsem2
    [/= hty1 hsync1 hsync2 [/= ??? hst]];
    subst.
  move: hsem1 hsem2 => /stepI [/= ??] /stepI [/= ??]; subst.
  split=> //.
  move: hty1 => [/= theta [hty1 hcs]].
  move: hty1 =>
    /typedcI [sigma0 [gamma0 [sigmai [gammai [sigma0' [gamma0' []]]]]]]
    /typediI [???] htyc [] ????;
    subst=> //.
  exists (MSFTupdated x), gamma0.
  split; last split=> //=.
  - by eeauto.
  - exact: (synced_after_update_msf (dt := {| dt_mem := st1; |})).
  - exact: (synced_after_update_msf (dt := {| dt_mem := st2; |})).
  by eauto using eq_dt_ctx_after_update_msf, weak_eq_dt.
Qed.

Lemma sss_protect st1 d x y msf c :
  st_c st1 = Iprotect x y msf :: c
  -> d = Dstep
  -> single_step_soundnessT st1 d.
Proof.
  move: st1 => [/= _ g cs [vm1 st1 ms]] -> ?; subst.
  move=> st1' [/= c2 g2 cs2 [vm2 st2 ms2]] st2' o1 o2 sigma gamma hsem1 hsem2
    [/= hty1 hsync1 hsync2 [/= ??? hst]];
    subst.
  move: hsem1 hsem2 => /stepI [/= ??] /stepI [/= ??]; subst.
  split=> //.
  move: hty1 => [/= theta [hty1 hcs]].
  move: hty1 =>
    /typedcI [sigma0 [gamma0 [sigmai [gammai [sigma0' [gamma0' []]]]]]]
    /typediI [???] htyc [] ????;
    subst=> //.
  exists (msft_restrict (MSFTupdated msf) x), (ctx_after_protect gamma0 x y).
  split; last split=> //=.
  - by eeauto.
  - exact: synced_msft_restrict.
  - exact: synced_msft_restrict.
  by eauto using eq_dt_ctx_after_protect, weak_eq_dt.
Qed.

Lemma sss_if st1 d b e c0 c1 c :
  st_c st1 = Iif e c0 c1 :: c
  -> d = Dforce b
  -> single_step_soundnessT st1 d.
Proof.
  move: st1 => [/= _ g cs [vm1 st1 ms]] -> ?; subst.
  move=> st1' [/= c2 g2 cs2 [vm2 st2 ms2]] st2' o1 o2 sigma gamma hsem1 hsem2
    [/= hty1 hsync1 hsync2 [/= ??? hst]];
    subst.
  move: hsem1 hsem2 => /stepI [/= b1' [hb ??]] /stepI [/= b2' [hb' ??]];
    subst.
  move: hty1 => [/= theta [hty1 hcs]].
  move: hty1 =>
    /typedcI [sigma0 [gamma0 [sigmai [gammai [sigma0' [gamma0' []]]]]]]
    /typediI [htye htyc0 htyc1] htyc ????.
  move: (hb).
  rewrite (eq_vm_eval_e_public (is_n := false) (vm' := vm2) _ htye erefl);
    first last.
  - move: hst => [[/= ? _] _ _]. by eauto using weak_eq_vm.
  rewrite hb' => -[?]; subst.
  split=> //.
  exists (msft_cond sigma0 (if b then e else enot e)), gamma0.
  split; last split=> //=.
  - exists theta; split=> //=.
    apply: typed_weak; first (by eauto using typed_after_if); by eeauto.

  - by apply: synced_msft_cond => //=; apply: (weak_synced hsync1).
  - by apply: synced_msft_cond => //=; apply: (weak_synced hsync2).
  apply: weak_eq_dt; last eassumption.
  by apply: eq_dt_with_ms hst.
Qed.

Lemma sss_while st1 d b e c0 c :
  st_c st1 = Iwhile e c0 :: c
  -> d = Dforce b
  -> single_step_soundnessT st1 d.
Proof.
  move: st1 => [/= _ g cs [vm1 st1 ms]] -> ?; subst.
  move=> st1' [/= c2 g2 cs2 [vm2 st2 ms2]] st2' o1 o2 sigma gamma hsem1 hsem2
    [/= hty1 hsync1 hsync2 [/= ??? hst]];
    subst.
  move: hsem1 hsem2 => /stepI [/= b1' [hb ??]] /stepI [/= b2' [hb' ??]];
    subst.
  move: hty1 => [/= theta [hty1 hcs]].
  move: hty1 =>
    /typedcI [sigma0 [gamma0 [? [? [? [? []]]]]]]
    /typediI [htye htyc0 ??] htyc ????;
    subst.
  move: (hb).
  rewrite (eq_vm_eval_e_public (is_n := false) (vm' := vm2) _ htye erefl);
    first last.
  - move: hst => [[/= ? _] _ _]. by eauto using weak_eq_vm.
  rewrite hb' => -[?]; subst.
  split=> //.
  exists (msft_cond sigma0 (if b then e else enot e)), gamma0.
  split; last split=> //=.
  - rewrite -cat_rcons -cats1 -{3}(cat0s c) -(fun_if (fun x => x ++ c)).
    exists theta; split=> //=.
    {
      apply: typed_weak.
      - apply: typed_after_if; [|by constructor | eassumption].
        apply: typed_capp; first eassumption.
        apply: typed_weak; by eeauto.
      all: by eeauto.
    }
  - apply: synced_msft_cond; by eauto using weak_synced.
  - apply: synced_msft_cond; by eauto using weak_synced.
  apply: weak_eq_dt; last eassumption.
  by apply: eq_dt_with_ms hst.
Qed.

Lemma sss_call st1 d l xi f c :
  st_c st1 = Icall l xi f :: c
  -> d = Dstep
  -> single_step_soundnessT st1 d.
Proof.
  move: st1 => [/= _ g cs [vm1 st1 ms]] -> ?; subst.
  move=> st1' [/= c2 g2 cs2 [vm2 st2 ms2]] st2' o1 o2 sigma gamma hsem1 hsem2
    [hty1 /= hsync1 hsync2 [/= ??? hst]];
    subst.
  move: hsem1 hsem2 => /stepI [/= ??] /stepI [/= ??]; subst.

  have := htyp f.
  set sigmaf := cert_msft (cert f).
  set sigmaf' := cert_msft' (cert f).
  set gammaf := cert_ctx (cert f).
  set gammaf' := cert_ctx' (cert f).
  move=> htyf.

  split=> //.
  move: hty1 => [/= thetag [hty1 hcs]].
  move: hty1 =>
    /typedcI [_ [_ [? [_ [? [? []]]]]]]
    /typediI [-> ? [thetaf [-> ->]]] htyc ????.
  exists sigmaf, (ctx_instantiate thetaf gammaf).
  split; last split=> //=.
  - exists thetaf; split=> /=; first exact: ctx_instantiate_typedc.
    by eeauto.
  - by eauto using weak_synced.
  - by eauto using weak_synced.
  by eauto using weak_eq_dt.
Qed.


Lemma sss_ret st1 d f l xi cont :
  st_c st1 = [::]
  -> d = Dreturn f l xi cont
  -> single_step_soundnessT st1 d.
Proof.
  move: st1 => [/= _ g cs [vm1 st1 ms]] -> ?; subst d.
  move=> st1' [/= c2 g2 cs2 [vm2 st2 ms2]] st2' o1 o2 sigma gamma hsem1 hsem2
    [/= hty1 hsync1 hsync2 [/= ??? [/= hspec ? hst]]];
    subst c2 g2 cs2 ms2.
  move: hsem1 hsem2 => /stepI /= [] ? hmspec /stepI /= [] ? hmspec'; subst o1 o2.
  split => //.
  move: hmspec hmspec'.
  case: (boolP (is_speculative_ret cs f l cont)) => hisspec.
  + move=> [hcont ->] [_ ->].
    move: hty1 => [/= theta [hty1 _]].
    move: hty1 => /typedcI [hsigma0 hgamma0].
    set sigmag' := cert_msft' (cert g).
    set gammag' := cert_ctx' (cert g).

    have := htyp f.
    set sigmaf' := cert_msft' (cert f).
    set gammaf' := cert_ctx' (cert f).
    move=> htyf.

    have {}hcont : List.In (l, xi, cont) (contc g (p f)).
    - by eauto using in_contpI.

    have [theta' [sigma_call [hsigma_call htycont]]] := typedc_contc htyf hcont.
    clear hcont htyf.

    exists sigma_call, (ctx_instantiate theta' gammag').
    split; last split=> //=.
    - exists theta_Secret; split=> /=; last by constructor.
      apply: typed_weak; last exact: ctx_instantiate_theta_Secret; by eeauto.
    - exact: (synced_after_ret_s (sigma := sigmag') {| dt_vm := _; |}).
    - exact: (synced_after_ret_s (sigma := sigmag') {| dt_vm := _; |}).
    split=> //=; last by constructor.
    apply: eq_dt_on_s_instantiate; by eauto.
  move=> -> ->.
  have heqcs : cs = (f, l, cont) :: drop 1 cs.
  - case: (cs) hisspec => //= -[[f' l'] cont'] ?.
    rewrite /is_speculative_ret /=.
    by rewrite negbK drop0 => /eqP ->.
  move: hty1 => [/= thetag [hty1 hcs]].
  move: hty1 => /typedcI [hmsftle hctxle].
  move: hcs; rewrite heqcs => /typed_callstackI [thetaf [htyc hcs]].
  exists (cert_msft' (cert g)), (ctx_instantiate thetag (cert_ctx' (cert g))).
  split; last split=> //=.
  - rewrite /= drop0. by eeauto.
  - by eauto using weak_synced.
  - by eauto using weak_synced.
  by apply: weak_eq_dt hctxle; split.
Qed.

Lemma main st1 d :
  single_step_soundnessT st1 d.
Proof.
  move=> > [] > *.
  - apply: sss_assign; by eeauto.
  - apply: sss_load_n; by eeauto.
  - apply: sss_load_s; by eeauto.
  - apply: sss_store_n; by eeauto.
  - apply: sss_store_s; by eeauto.
  - apply: sss_init_msf; by eeauto.
  - apply: sss_update_msf; by eeauto.
  - apply: sss_protect; by eeauto.
  - apply: sss_if; by eeauto.
  - apply: sss_while; by eeauto.
  - apply: sss_call; by eeauto.
  apply: sss_ret; by eeauto.
Qed.

End PROOF.

End __SingleStepSoundness.

Definition single_step_soundness := __SingleStepSoundness.main.


(* -------------------------------------------------------------------------- *)
(* Multi-step soundness. *)

Section MULTISTEP.

Context
  (cert : certificate)
  (p : program)
  (htyp : typedp cert p)
.

Notation typedi := (typedi cert).
Notation typedc := (typedc cert).
Notation step := (step p).
Notation sem := (sem p).
Notation typedm := (typedm cert).

Lemma multi_step_soundness st1 st1' st2 st2' ds os1 os2 sigma gamma :
  let: dt1 := st_dt st1 in
  let: dt2 := st_dt st2 in
  let: dt1' := st_dt st1' in
  let: dt2' := st_dt st2' in
  sem st1 ds os1 st1'
  -> sem st2 ds os2 st2'
  -> typedm sigma gamma st1
  -> synced sigma (dt_vm dt1) (dt_ms dt1)
  -> synced sigma (dt_vm dt2) (dt_ms dt2)
  -> eq_st gamma st1 st2
  -> os1 = os2
     /\ exists sigma' gamma',
          [/\ typedm sigma' gamma' st1'
            , synced sigma' (dt_vm dt1') (dt_ms dt1')
            , synced sigma' (dt_vm dt2') (dt_ms dt2')
            & eq_st gamma' st1' st2'
          ].
Proof.
  elim: ds st1 st1' st2 st2' os1 os2 sigma gamma
    => [|d ds hind] st1 st1' st2 st2' os1 os2 sigma gamma
    /semI hsem1 /semI hsem2 htym1 hsync1 hsync2 heqm.
  - move: hsem1 hsem2 => [??] [??]; subst. by eeauto.
  move: hsem1 hsem2
    => [o1 [os1' [? [mi1 hstep1 hsem1]]]] [o2 [os2' [? [mi2 hstep2 hsem2]]]];
    subst.
  have hsss_invariant := And4 htym1 hsync1 hsync2 heqm.
  have [? [sigmai [gammai [htymi1 hsynci1 hsynci2 heqmi]]]] :=
    single_step_soundness htyp hstep1 hstep2 hsss_invariant.
  subst.
  by have [-> ?] :=
    hind _ _ _ _ _ _ _ _ hsem1 hsem2 htymi1 hsynci1 hsynci2 heqmi.
Qed.

Lemma typedm_initial_state st :
  let: c := cert (p_entry_fn p) in
  let: sigma := cert_msft c in
  let: gamma := cert_ctx c in
  typedp cert p
  -> is_initial_state p st
  -> typedm sigma gamma st.
Proof.
  move: st => [c f cs dt] /(_ (p_entry_fn p)).
  move=> hty /and3P [] /= /eqP ? /eqP ? /eqP ?; subst c f cs.
  exists theta_Secret; split; last by constructor.
  apply: typed_weak; last exact: ctx_instantiate_theta_Secret; by eeauto.
Qed.

Lemma eqm_initial_state st1 st2 :
  let: c := cert (p_entry_fn p) in
  let: gamma := cert_ctx c in
  let: dt1 := st_dt st1 in
  let: dt2 := st_dt st2 in
  initial_phi p gamma st1 st2
  -> eq_st gamma st1 st2.
Proof.
  move=> [] /and3P [] /eqP ? /eqP ? /eqP ? /and3P [] /eqP ? /eqP ? /eqP ? []
    [??] [??] ?.
  split; [congruence | congruence | congruence|].
  split=> //.
  by right.
Qed.

End MULTISTEP.

Lemma typedp_soundness p cert :
  Typedp_soundness p cert.
Proof.
  set c := cert (p_entry_fn p).
  set gamma := cert_ctx c.
  move=> /= hsigma htyp ds os1 os2 st1 st2 st1' st2' [hm1 hm2 hst] hsem1 hsem2.
  have htym1 := typedm_initial_state htyp hm1.
  have htym2 := typedm_initial_state htyp hm2.
  have heqm := eqm_initial_state (And3 hm1 hm2 hst).
  have [||-> //] := multi_step_soundness htyp hsem1 hsem2 htym1 _ _ heqm.
  all: by move: hsigma => /wfcertP ->.
Qed.
