(* -------------------------------------------------------------------------- *)
(* Semantics. *)
(* -------------------------------------------------------------------------- *)

From Coq Require Import ZArith.
From Coq Require Relations.
From mathcomp Require Import all_ssreflect.

Require Import
  syntax
  utils
  var
.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Open Scope Z.


(* -------------------------------------------------------------------------- *)
(* Values. *)

Variant value :=
  | Vint of Z
  | Vbool of bool
  | Vundef
.

(* These can be anything, we only need them to be different. *)
Definition msf_mask : value := Vbool true.
Definition msf_nomask : value := Vbool false.

Scheme Equality for value.

Lemma value_eq_axiom : Equality.axiom value_beq.
Proof.
  exact: (eq_axiom_of_scheme internal_value_dec_bl internal_value_dec_lb).
Qed.

Definition value_eqMixin := Equality.Mixin value_eq_axiom.
Canonical value_eqType := EqType value value_eqMixin.


(* -------------------------------------------------------------------------- *)
(* Variable maps. *)

Module Type VmT.
  Parameter t : Type.
  Parameter get : t -> regvar -> value.
  Parameter set : t -> regvar -> value -> t.

  Parameter get_set :
    forall vm x y v,
      get (set vm x v) y = if x == y then v else get vm y.
End VmT.

Module Vm : VmT.
  Definition t := regvar -> value.

  Definition get (vm : t) (x : regvar) : value  := vm x.
  Definition set (vm : t) (x : regvar) (v : value) : t :=
    fun y => if x == y then v else vm y.

  Lemma get_set vm x y v :
    get (set vm x v) y = if x == y then v else get vm y.
  Proof. done. Qed.
End Vm.


(* -------------------------------------------------------------------------- *)
(* Memory. *)
(* We have a segmented memory (in arrays) and reads or writes that are out of
   bounds always return [Vundef].
   We still capture the attacker's capabilities because the rules for unsafe
   loads and stores allows them to load or store any value from any array. *)

Definition in_bounds (a : arrvar) (i : Z) :=
  [&& 0 <=? i & i <? ArrayVar.len a ]%Z.

Module Type MemT.
  Parameter t : Type.
  Parameter read : t -> arrvar -> Z -> value.
  Parameter write : t  -> arrvar -> Z -> value -> t.

  Parameter read_write :
    forall st a i b j v,
      in_bounds a i
      -> read (write st a i v) b j
         = if [&& a == b & Z.eqb i j ] then v else read st b j.

End MemT.

Module Mem : MemT.
  Record _t :=
    {
      mem_map : arrvar -> Z -> value;
    }.

  Definition t := _t.
  Definition undef : t :=
    {| mem_map := fun _ _ => Vundef; |}.

  Definition read (st : t) (a : arrvar) (i : Z) : value :=
    if in_bounds a i then mem_map st a i else Vundef.

  Definition write (st : t) (a : arrvar) (i : Z) (v : value) : t :=
    if in_bounds a i
    then
      let map' a' i' :=
        if [&& a' == a & Z.eqb i i' ] then v else mem_map st a' i'
      in
      {| mem_map := map'; |}
    else
      undef.

  Lemma read_write st a i b j v :
    in_bounds a i
    -> read (write st a i v) b j
       = if [&& a == b & Z.eqb i j ] then v else read st b j.
  Proof.
    move=> hi.
    rewrite /write hi /read /= /in_bounds /= -/(in_bounds b j) eq_sym.
    case: andP; last done.
    move=> [/eqP ? /Z.eqb_eq ?]; subst b j.
    by rewrite hi.
  Qed.

End Mem.

(* -------------------------------------------------------------------------- *)
(* States. *)

Record data :=
  {
    dt_vm : Vm.t;
    dt_mem : Mem.t;
    dt_ms : bool;
  }.

Definition with_vm dt vm :=
  {|
    dt_vm := vm;
    dt_mem := dt_mem dt;
    dt_ms := dt_ms dt;
  |}.

Definition with_mem dt mem :=
  {|
    dt_vm := dt_vm dt;
    dt_mem := mem;
    dt_ms := dt_ms dt;
  |}.

Definition with_ms dt ms :=
  {|
    dt_vm := dt_vm dt;
    dt_mem := dt_mem dt;
    dt_ms := ms;
  |}.

Definition set_ms dt := with_ms dt true.


(* -------------------------------------------------------------------------- *)
(* Evaluating expressions. *)

Fixpoint eval (vm : Vm.t) (e : expression) : value :=
  match e with
  | Econst z => Vint z
  | Ebool b => Vbool b
  | Evar x => Vm.get vm x
  | Eop1 Onot e => if eval vm e is Vbool b then Vbool (~~ b) else Vundef
  | Eop1 Oneg e => if eval vm e is Vint z then Vint (- z)%Z else Vundef
  | Eop2 op e0 e1 =>
      match op, eval vm e0, eval vm e1 with
      | Oadd, Vint z0, Vint z1 => Vint (z0 + z1)
      | Oeq, Vint z0, Vint z1 => Vbool (z0 == z1)%Z
      | Olt, Vint z0, Vint z1 => Vbool (z0 <? z1)%Z
      | Oeq, Vbool b0, Vbool b1 => Vbool (b0 == b1)
      | Oand, Vbool b0, Vbool b1 => Vbool (b0 && b1)
      | _, _, _ => Vundef
      end
  end.


(* -------------------------------------------------------------------------- *)
(* Writing. *)

Definition write_reg (dt : data) (x : regvar) (v : value) : data :=
  Vm.set (dt_vm dt) x v |> with_vm dt.

Definition write_mem (dt : data) (a : arrvar) (i : Z) (v : value) : data :=
  Mem.write (dt_mem dt) a i v |> with_mem dt.


(* -------------------------------------------------------------------------- *)
(* Machines. *)

Definition callstack := seq (funname * label * code).

Record state :=
  {
    st_c : code;
    st_fn : funname;
    st_cs : callstack;
    st_dt : data;
  }.

Definition with_c (st : state) (c : code) : state :=
  {|
    st_c := c;
    st_fn := st_fn st;
    st_cs := st_cs st;
    st_dt := st_dt st;
  |}.

Definition with_dt (st : state) (dt : data) : state :=
  {|
    st_c := st_c st;
    st_fn := st_fn st;
    st_cs := st_cs st;
    st_dt := dt;
  |}.


(* -------------------------------------------------------------------------- *)
(* Small step semantics. *)

Variant directive :=
  | Dstep
  | Dforce of bool
  | Dmem of arrvar & Z
  | Dreturn of funname & label & option regvar & code
.

Variant observation :=
  | Onone
  | Oaddr of arrvar & Z
  | Obranch of bool
.

Definition st_after_assign st c x e :=
  let dt := st_dt st in
  let v := eval (dt_vm dt) e in
  let dt' := write_reg dt x v in
  with_c (with_dt st dt') c.

Definition st_after_init_msf st c msf :=
  let dt' := write_reg (st_dt st) msf msf_nomask in
  with_c (with_dt st dt') c.

Definition dt_after_update_msf dt x y e :=
  let vm := dt_vm dt in
  let v := if eval vm e == Vbool true then Vm.get vm y else msf_mask in
  write_reg dt x v.

Definition st_after_update_msf st c x y e :=
  let dt' := dt_after_update_msf (st_dt st) x y e in
  with_c (with_dt st dt') c.

Definition dt_after_protect dt x y msf :=
  let vm := dt_vm dt in
  let v := if Vm.get vm msf == msf_mask then msf_mask else Vm.get vm y in
  write_reg dt x v.

Definition st_after_protect st c x y msf :=
  let dt' := dt_after_protect (st_dt st) x y msf in
  with_c (with_dt st dt') c.

Definition st_after_load st c x a i :=
  let dt := st_dt st in
  let v := Mem.read (dt_mem dt) a i in
  let dt' := write_reg dt x v in
  with_c (with_dt st dt') c.

Definition st_after_store st c a i x :=
  let dt := st_dt st in
  let v := Vm.get (dt_vm dt) x in
  let dt' := write_mem dt a i v in
  with_c (with_dt st dt') c.

Definition st_after_call st c f cf l :=
  {|
    st_c := cf;
    st_fn := f;
    st_cs := (st_fn st, l, c) :: st_cs st;
    st_dt := st_dt st;
  |}.

Definition st_after_ret_n st f c cs :=
  {|
    st_c := c;
    st_fn := f;
    st_cs := cs;
    st_dt := st_dt st;
  |}.

Definition dt_after_RET_S dt xi :=
  let vm := dt_vm dt in
  let vm' := if xi is Some x then Vm.set vm x msf_mask else vm in
  with_vm dt vm' |> set_ms.

Definition st_after_ret_s st f xi c :=
  {|
    st_c := c;
    st_fn := f;
    st_cs := [::];
    st_dt := dt_after_RET_S (st_dt st) xi;
  |}.

Definition is_speculative_ret
  (cs : callstack) (f : funname) (l : label) (c : code) : bool :=
  match ohead cs with
  | Some (f', l', c') => (f, l, c) != (f', l', c')
  | None => true
  end.

(* The rules for unsafe loads and stores require that the misspeculation status
   is set, because we assume that the program is sequentially safe, and
   therefore unsafe memory accesses must happen under misspeculation. *)
Variant step
  (p : program) : state -> directive -> observation -> state -> Prop :=
  | step_assign :
    forall st x e c,
      st_c st = Iassign x e :: c
      -> step p st Dstep Onone (st_after_assign st c x e)

  | step_load_n :
    forall st x a e c i,
      let: dt := st_dt st in
      let: st' := st_after_load st c x a i in
      st_c st = Iload x a e :: c
      -> eval (dt_vm dt) e = Vint i
      -> in_bounds a i
      -> step p st Dstep (Oaddr a i) st'

  | step_load_s :
    forall st x a e c i b j,
      let: dt := st_dt st in
      let: st' := st_after_load st c x b j in
      st_c st = Iload x a e :: c
      -> eval (dt_vm dt) e = Vint i
      -> ~~ in_bounds a i
      -> in_bounds b j
      -> dt_ms dt
      -> step p st (Dmem b j) (Oaddr a i) st'

  | step_store_n :
    forall st a e x c i,
      let: dt := st_dt st in
      let: st' := st_after_store st c a i x in
      st_c st = Istore a e x :: c
      -> eval (dt_vm dt) e = Vint i
      -> in_bounds a i
      -> step p st Dstep (Oaddr a i) st'

  | step_store_s :
    forall st a e x c i b j,
      let: dt := st_dt st in
      let: st' := st_after_store st c b j x in
      st_c st = Istore a e x :: c
      -> eval (dt_vm dt) e = Vint i
      -> ~~ in_bounds a i
      -> in_bounds b j
      -> dt_ms dt
      -> step p st (Dmem b j) (Oaddr a i) st'

  | step_init_msf :
    forall st msf c,
      st_c st = Iinit_msf msf :: c
      -> ~~ dt_ms (st_dt st)
      -> step p st Dstep Onone (st_after_init_msf st c msf)

  | step_update_msf :
    forall st x y e c,
      st_c st = Iupdate_msf x y e :: c
      -> step p st Dstep Onone (st_after_update_msf st c x y e)

  | step_protect :
    forall st x y msf c,
      st_c st = Iprotect x y msf :: c
      -> step p st Dstep Onone (st_after_protect st c x y msf)

  | step_if :
    forall st e c1 c0 c b b',
      st_c st = Iif e c1 c0 :: c
      -> let: dt := st_dt st in
         eval (dt_vm dt) e = Vbool b'
         -> let: c' := (if b then c1 else c0) ++ c in
            let: dt' := with_ms dt ((b != b') || dt_ms dt) in
            let: st' := with_c (with_dt st dt') c' in
            step p st (Dforce b) (Obranch b') st'

  | step_while :
    forall st e c1 c b b',
      st_c st = Iwhile e c1 :: c
      -> let: dt := st_dt st in
         eval (dt_vm dt) e = Vbool b'
         -> let: c' := if b then c1 ++ (st_c st) else c in
            let: dt' := with_ms dt ((b != b') || dt_ms dt) in
            let: st' := with_c (with_dt st dt') c' in
            step p st (Dforce b) (Obranch b') st'

  | step_call :
    forall st f l xi c,
      st_c st = Icall l xi f :: c
      -> step p st Dstep Onone (st_after_call st c f (p f) l)

  | step_ret :
    forall st f l xi c st',
      st_c st = [::]
      -> (if is_speculative_ret (st_cs st) f l c
          then [/\ List.In (f, l, xi, c) (contp (st_fn st) p)
                 & st' = st_after_ret_s st f xi c
               ]
          else st' = st_after_ret_n st f c (drop 1 (st_cs st)))
      -> step p st (Dreturn f l xi c) Onone st'.

Inductive sem
  (p : program) :
  state -> seq directive -> seq observation -> state -> Prop :=
  | sem_done :
    forall st, sem p st [::] [::] st

  | sem_step :
    forall st st' st'' d o ds os,
      step p st d o st'
      -> sem p st' ds os st''
      -> sem p st (d :: ds) (o :: os) st''
.

Definition is_initial_state (p : program) (st : state) : bool :=
  [&& st_c st == p (p_entry_fn p)
    , st_fn st == p_entry_fn p
    & st_cs st == [::]
  ].
