From Coq Require Import ZArith.
From mathcomp Require Import all_ssreflect.

Require Import
  semantics
  syntax
  syntax_facts
  utils
  utils_facts
.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Lemma with_ms_id dt :
  with_ms dt (dt_ms dt) = dt.
Proof. by move: dt => []. Qed.

Lemma set_eq vm x v :
  Vm.get (Vm.set vm x v) x = v.
Proof. by rewrite Vm.get_set eqxx. Qed.

Lemma set_neq vm x y v :
  x != y
  -> Vm.get (Vm.set vm x v) y = Vm.get vm y.
Proof. by rewrite Vm.get_set => /negPf ->. Qed.

Lemma eval_set vm e x v' :
  ~~ Sr.mem x (free_variables e)
  -> eval (Vm.set vm x v') e = eval vm e.
Proof.
  elim: e => [// | // | y | op e hind | op e0 hind0 e1 hind1].
  - rewrite Sr.F.singleton_b Sr.eqbP eq_sym => ?. by rewrite /= set_neq.
  - case: op => /= hx; by erewrite hind.
  rewrite Sr.F.union_b -/free_variables.
  case: op => /= /norP [??]; by rewrite hind0 // hind1.
Qed.

Definition eq_on (s : Sr.t) (vm vm' : Vm.t) : Prop :=
  forall x, Sr.mem x s -> Vm.get vm x = Vm.get vm' x.

Lemma eq_on_cat s0 s1 vm vm' :
  eq_on (Sr.union s0 s1) vm vm'
  -> eq_on s0 vm vm' /\ eq_on s1 vm vm'.
Proof.
  move=> h. split=> y hy; apply: h; by rewrite Sr.F.union_b hy // orbT.
Qed.

Lemma eval_eq_on vm vm' e :
  eq_on (free_variables e) vm vm'
  -> eval vm e = eval vm' e.
Proof.
  elim: e => [// | // | x | op e hind | op e0 hind0 e1 hind1] /=.
  - move=> -> //. by rewrite Sr.F.singleton_b Sr.eqbP eqxx.
  - by move=> /hind ->.
  by move=> /eq_on_cat [] /hind0 -> /hind1 ->.
Qed.

Lemma stepI p st d o st' :
  let: dt := st_dt st in
  step p st d o st'
  -> match d, st_c st with
     | Dstep, Iassign x e :: c =>
         [/\ o = Onone
           & st' = st_after_assign st c x e
         ]

     | Dstep, Iload x a e :: c =>
         exists i,
           [/\ eval (dt_vm dt) e = Vint i
             , in_bounds a i
             , o = Oaddr a i
             & st' = st_after_load st c x a i
           ]

     | Dmem b j, Iload x a e :: c =>
         exists i,
           [/\ eval (dt_vm dt) e = Vint i
             , ~~ in_bounds a i
             , in_bounds b j
             , dt_ms dt
             , o = Oaddr a i
             & st' = st_after_load st c x b j
           ]

     | Dstep, Istore a e x :: c =>
         exists i,
           [/\ eval (dt_vm dt) e = Vint i
             , in_bounds a i
             , o = Oaddr a i
             & st' = st_after_store st c a i x
           ]

     | Dmem b j, Istore a e x :: c =>
         exists i,
           [/\ eval (dt_vm dt) e = Vint i
             , ~~ in_bounds a i
             , in_bounds b j
             , dt_ms dt
             , o = Oaddr a i
             & st' = st_after_store st c b j x
           ]

     | Dstep, Iinit_msf msf :: c =>
         [/\ ~~ dt_ms (st_dt st)
           , o = Onone
           & st' = st_after_init_msf st c msf
         ]

     | Dstep, Iupdate_msf x y e :: c =>
         [/\ o = Onone
           & st' = st_after_update_msf st c x y e
         ]

     | Dstep, Iprotect x y msf :: c =>
         [/\ o = Onone
           & st' = st_after_protect st c x y msf
         ]

     | Dforce b, Iif e c0 c1 :: c =>
         exists b',
           let: dt := st_dt st in
           [/\ eval (dt_vm dt) e = Vbool b'
             , o = Obranch b'
             & let: c' := (if b then c0 else c1) ++ c in
               let: dt' := with_ms dt ((b != b') || dt_ms dt) in
               st' = with_dt (with_c st c') dt'
           ]

     | Dforce b, Iwhile e c0 :: c =>
         exists b',
           let: dt := st_dt st in
           [/\ eval (dt_vm (st_dt st)) e = Vbool b'
             , o = Obranch b'
             & let: c' := if b then c0 ++ (st_c st) else c in
               let: dt' := with_ms dt ((b != b') || dt_ms dt) in
               st' = with_dt (with_c st c') dt'
           ]

     | Dstep, Icall l xi f :: c =>
         [/\ o = Onone
           & st' = st_after_call st c f (p f) l
         ]

     | Dreturn f l xi c, [::] =>
         [/\ o = Onone
           & if is_speculative_ret (st_cs st) f l c
             then
               [/\ List.In (f, l, xi, c) (contp (st_fn st) p)
                 & st' = st_after_ret_s st f xi c
               ]
             else st' = st_after_ret_n st f c (drop 1 (st_cs st))
         ]

     | _, _ => False
     end.
Proof. case=> -[??? [???]] /= *; subst; by eeauto. Qed.

Lemma semI p st ds os st' :
  sem p st ds os st'
  -> match ds with
     | [::] => st = st' /\ os = [::]
     | d :: ds' =>
         exists o os',
           os = o :: os' /\ exists2 mi, step p st d o mi & sem p mi ds' os' st'
     end.
Proof. case; by eeauto. Qed.

Lemma sem_step1 p st d o st' :
  step p st d o st' ->
  sem p st [:: d ] [:: o ] st'.
Proof. by eeauto. Qed.

Lemma sem_cat p st st' st'' D D' O O' : 
  sem p st D O st' ->
  sem p st' D' O' st'' ->
  sem p st (D ++ D') (O ++ O') st''.
Proof. by elim => //= > hstep _ h /h; econstructor; eauto. Qed.

Lemma size_obs p st D O st' : sem p st D O st' -> size D = size O.
Proof. by elim => //= > _ _ ->. Qed.
