/*
 * Copyright (c) 2024-2025 The mlkem-native project authors
 * SPDX-License-Identifier: Apache-2.0
 */

// AArch64 re-implementation of the asymmetric base multiplication from:

// Neon NTT: Faster Dilithium, Kyber, and Saber on Cortex-A72 and Apple M1
// https://eprint.iacr.org/2021/986
// https://github.com/neon-ntt/neon-ntt

#include "../../../common.h"
#if defined(MLKEM_NATIVE_ARITH_BACKEND_AARCH64_OPT) && \
    !defined(MLKEM_NATIVE_MULTILEVEL_BUILD_NO_SHARED) && \
    (defined(MLKEM_NATIVE_MULTILEVEL_BUILD_WITH_SHARED) || MLKEM_K == 2)
/* simpasm: header-end */

// Input:
// - Vectors al, ah of 32-bit entries
// Output:
// - Montgomery reductions of al || ah, stored in al
.macro montgomery_reduce_long x, a
        uzp1   t0.8h, \a\()l.8h, \a\()h.8h
        mul    t0.8h, t0.8h, modulus_twisted.8h
        smlal  \a\()l.4s, t0.4h, modulus.4h
        smlal2 \a\()h.4s, t0.8h, modulus.8h
        uzp2   \x\().8h, \a\()l.8h, \a\()h.8h
.endm

// Computes products (a0*b0 + a0*b0t, a0*b1 + a1*b0) in 32-bit.

// Bounds:
// - Assume |a| < 4096,
// - Result: < 2*4096*2^15 = 2^28
.macro pmull d, a, b
        smull  \d\()0l.4s, \a\()0.4h, \b\()0.4h
        smull2 \d\()0h.4s, \a\()0.8h, \b\()0.8h
        smlal  \d\()0l.4s, \a\()1.4h, \b\()1t.4h
        smlal2 \d\()0h.4s, \a\()1.8h, \b\()1t.8h

        smull  \d\()1l.4s, \a\()0.4h, \b\()1.4h
        smull2 \d\()1h.4s, \a\()0.8h, \b\()1.8h
        smlal  \d\()1l.4s, \a\()1.4h, \b\()0.4h
        smlal2 \d\()1h.4s, \a\()1.8h, \b\()0.8h
.endm

.macro pmlal d, a, b
        smlal  \d\()0l.4s, \a\()0.4h, \b\()0.4h
        smlal2 \d\()0h.4s, \a\()0.8h, \b\()0.8h
        smlal  \d\()0l.4s, \a\()1.4h, \b\()1t.4h
        smlal2 \d\()0h.4s, \a\()1.8h, \b\()1t.8h

        smlal  \d\()1l.4s, \a\()0.4h, \b\()1.4h
        smlal2 \d\()1h.4s, \a\()0.8h, \b\()1.8h
        smlal  \d\()1l.4s, \a\()1.4h, \b\()0.4h
        smlal2 \d\()1h.4s, \a\()1.8h, \b\()0.8h
.endm

.macro ld2_wrap a, ptr
        ldr q_tmp0, [\ptr\()], #32
        ldr q_tmp1, [\ptr\(), #-16]
        uzp1 \a\()0.8h, tmp0.8h, tmp1.8h
        uzp2 \a\()1.8h, tmp0.8h, tmp1.8h
.endm

.macro st2_wrap a, ptr
        zip1 tmp0.8h, \a\()0.8h, \a\()1.8h
        zip2 tmp1.8h, \a\()0.8h, \a\()1.8h
        str q_tmp0, [\ptr\()], #32
        str q_tmp1, [\ptr\(), #-16]
.endm

.macro load_polys a, b, a_ptr, b_ptr, b_cache_ptr
        ld2_wrap \a\(), \a_ptr
        ld2_wrap \b\(), \b_ptr
        ld1 {\b\()1t.8h}, [\b_cache_ptr], #16
.endm

.macro save_vregs
        sub sp, sp, #(16*4)
        stp  d8,  d9, [sp, #16*0]
        stp d10, d11, [sp, #16*1]
        stp d12, d13, [sp, #16*2]
        stp d14, d15, [sp, #16*3]
.endm

.macro restore_vregs
        ldp  d8,  d9, [sp, #16*0]
        ldp d10, d11, [sp, #16*1]
        ldp d12, d13, [sp, #16*2]
        ldp d14, d15, [sp, #16*3]
        add sp, sp, #(16*4)
.endm

.macro push_stack
        save_vregs
.endm

.macro pop_stack
        restore_vregs
.endm

        out          .req x0
        a0_ptr       .req x1
        b0_ptr       .req x2
        b0_cache_ptr .req x3
        a1_ptr       .req x4
        b1_ptr       .req x5
        b1_cache_ptr .req x6
        a2_ptr       .req x7
        b2_ptr       .req x8
        b2_cache_ptr .req x9
        a3_ptr       .req x10
        b3_ptr       .req x11
        b3_cache_ptr .req x12
        count        .req x13
        wtmp         .req w14

        modulus           .req v0
        modulus_twisted   .req v2

        aa0      .req v3
        aa1      .req v4
        bb0      .req v5
        bb1      .req v6
        bb1t     .req v7

        res0l   .req v8
        res1l   .req v9
        res0h   .req v10
        res1h   .req v11

        tmp0 .req v12
        tmp1 .req v13
        q_tmp0 .req q12
        q_tmp1 .req q13

        out0 .req v26
        out1 .req v27

        t0   .req v28

        .text
        .global MLKEM_ASM_NAMESPACE(polyvec_basemul_acc_montgomery_cached_asm_k2_opt)
        .balign 4
MLKEM_ASM_NAMESPACE(polyvec_basemul_acc_montgomery_cached_asm_k2_opt):
        push_stack

        mov wtmp, #3329
        dup modulus.8h, wtmp

        mov wtmp, #3327
        dup modulus_twisted.8h, wtmp

        // Computed bases of vector entries

        add a1_ptr, a0_ptr, #(1 * 512)
        add b1_ptr, b0_ptr, #(1 * 512)
        add b1_cache_ptr, b0_cache_ptr, #(1 * 512/2)

        mov count, #(MLKEM_N / 16)
                                             // Instructions:    75
                                             // Expected cycles: 94
                                             // Expected IPC:    0.80

                                             // Cycle bound:     94.0
                                             // IPC bound:       0.80

                                             // Wall time:     1.49s
                                             // User time:     1.49s

                                             // --------------------------- original position ---------------------------->
                                             // 0                        25                       50
                                             // |------------------------|------------------------|
        ldr q9, [x4], #32                    // *..........................................................................
        ldr q5, [x4, #-16]                   // ......*....................................................................
        ldr q11, [x5], #32                   // .*.........................................................................
        uzp1 v23.8H, v9.8H, v5.8H            // .........*.................................................................
        uzp2 v9.8H, v9.8H, v5.8H             // .....................*.....................................................
        ldr q5, [x2], #32                    // ..*........................................................................
        ldr q7, [x5, #-16]                   // ..............*............................................................
        ldr q21, [x2, #-16]                  // ...*.......................................................................
        uzp2 v10.8H, v11.8H, v7.8H           // .................*.........................................................
        uzp1 v11.8H, v11.8H, v7.8H           // ..................*........................................................
        uzp1 v7.8H, v5.8H, v21.8H            // ....*......................................................................
        uzp2 v5.8H, v5.8H, v21.8H            // .....*.....................................................................
        ldr q21, [x1], #32                   // .......*...................................................................
        ldr q25, [x1, #-16]                  // ........*..................................................................
        ld1 {v6.8H}, [x3], #16               // ............................*..............................................
        uzp1 v26.8H, v21.8H, v25.8H          // ..........*................................................................
        uzp2 v21.8H, v21.8H, v25.8H          // ...........*...............................................................
        smull v25.4S, v26.4H, v5.4H          // ............*..............................................................
        smull2 v5.4S, v26.8H, v5.8H          // .............*.............................................................
        smull v19.4S, v26.4H, v7.4H          // ..........................*................................................
        smull2 v26.4S, v26.8H, v7.8H         // ..............................*............................................
        smlal v25.4S, v21.4H, v7.4H          // ...............*...........................................................
        smlal2 v5.4S, v21.8H, v7.8H          // ................*..........................................................
        smlal v19.4S, v21.4H, v6.4H          // ...................................*.......................................
        smlal2 v26.4S, v21.8H, v6.8H         // .................................*.........................................
        smlal v25.4S, v23.4H, v10.4H         // ...................*.......................................................
        smlal2 v5.4S, v23.8H, v10.8H         // ....................*......................................................
        smlal v19.4S, v23.4H, v11.4H         // ......................................*....................................
        smlal2 v26.4S, v23.8H, v11.8H        // ....................................*......................................
        ld1 {v23.8H}, [x6], #16              // ........................*..................................................
        smlal v25.4S, v9.4H, v11.4H          // ......................*....................................................
        smlal2 v5.4S, v9.8H, v11.8H          // .......................*...................................................
        smlal2 v26.4S, v9.8H, v23.8H         // .......................................*...................................
        smlal v19.4S, v9.4H, v23.4H          // .........................................*.................................
        ldr q9, [x4], #32                    // ...............................*...........................................
        uzp1 v11.8H, v25.8H, v5.8H           // .........................*.................................................
        uzp1 v23.8H, v19.8H, v26.8H          // .............................................*.............................
        mul v11.8H, v11.8H, v2.8H            // ...........................*...............................................
        mul v23.8H, v23.8H, v2.8H            // ..............................................*............................
        ldr q7, [x5], #32                    // ................................*..........................................
        smlal2 v5.4S, v11.8H, v0.8H          // .............................*.............................................
        smlal v25.4S, v11.4H, v0.4H          // ..................................*........................................
        ldr q11, [x2], #32                   // .....................................*.....................................
        ldr q21, [x2, #-16]                  // ........................................*..................................
        ldr q6, [x4, #-16]                   // ...............................................*...........................
        uzp1 v17.8H, v11.8H, v21.8H          // ...........................................*...............................
        ldr q10, [x1], #32                   // ................................................*..........................
        ldr q29, [x1, #-16]                  // .................................................*.........................
        uzp2 v11.8H, v11.8H, v21.8H          // ............................................*..............................
        uzp1 v13.8H, v9.8H, v6.8H            // ...................................................*.......................
        uzp1 v3.8H, v10.8H, v29.8H           // ....................................................*......................
        uzp2 v10.8H, v10.8H, v29.8H          // .....................................................*.....................
        smull v12.4S, v3.4H, v11.4H          // ......................................................*....................
        smull2 v11.4S, v3.8H, v11.8H         // .......................................................*...................
        ldr q21, [x5, #-16]                  // ........................................................*..................
        smlal v12.4S, v10.4H, v17.4H         // .........................................................*.................
        smlal2 v11.4S, v10.8H, v17.8H        // ..........................................................*................
        uzp2 v29.8H, v7.8H, v21.8H           // ...........................................................*...............
        uzp1 v15.8H, v7.8H, v21.8H           // ............................................................*..............
        smlal v12.4S, v13.4H, v29.4H         // .............................................................*.............
        smlal2 v11.4S, v13.8H, v29.8H        // ..............................................................*............
        uzp2 v28.8H, v9.8H, v6.8H            // ...............................................................*...........
        smlal2 v26.4S, v23.8H, v0.8H         // ..................................................*........................
        smlal v12.4S, v28.4H, v15.4H         // .................................................................*.........
        smlal2 v11.4S, v28.8H, v15.8H        // ..................................................................*........
        smlal v19.4S, v23.4H, v0.4H          // ................................................................*..........
        uzp2 v27.8H, v25.8H, v5.8H           // ..........................................*................................
        smull v23.4S, v3.4H, v17.4H          // ......................................................................*....
        uzp1 v9.8H, v12.8H, v11.8H           // .....................................................................*.....
        uzp2 v19.8H, v19.8H, v26.8H          // ....................................................................*......
        mul v14.8H, v9.8H, v2.8H             // .......................................................................*...
        ld1 {v22.8H}, [x6], #16              // ...................................................................*.......
        zip2 v9.8H, v19.8H, v27.8H           // ........................................................................*..
        smlal2 v11.4S, v14.8H, v0.8H         // ..........................................................................*
        ld1 {v4.8H}, [x3], #16               // .........................................................................*.

                                              // ------------------------------ new position ------------------------------>
                                              // 0                        25                       50
                                              // |------------------------|------------------------|------------------------
        // ldr q18, [x4], #32                   // *..........................................................................
        // ldr q30, [x5], #32                   // ..*........................................................................
        // ldr q8, [x2], #32                    // .....*.....................................................................
        // ldr q9, [x2, #-16]                   // .......*...................................................................
        // uzp1 v17.8H, v8.8H, v9.8H            // ..........*................................................................
        // uzp2 v4.8H, v8.8H, v9.8H             // ...........*...............................................................
        // ldr q19, [x4, #-16]                  // .*.........................................................................
        // ldr q29, [x1], #32                   // ............*..............................................................
        // ldr q12, [x1, #-16]                  // .............*.............................................................
        // uzp1 v13.8H, v18.8H, v19.8H          // ...*.......................................................................
        // uzp1 v3.8H, v29.8H, v12.8H           // ...............*...........................................................
        // uzp2 v10.8H, v29.8H, v12.8H          // ................*..........................................................
        // smull v12.4S, v3.4H, v4.4H           // .................*.........................................................
        // smull2 v11.4S, v3.8H, v4.8H          // ..................*........................................................
        // ldr q5, [x5, #-16]                   // ......*....................................................................
        // smlal v12.4S, v10.4H, v17.4H         // .....................*.....................................................
        // smlal2 v11.4S, v10.8H, v17.8H        // ......................*....................................................
        // uzp2 v14.8H, v30.8H, v5.8H           // ........*..................................................................
        // uzp1 v15.8H, v30.8H, v5.8H           // .........*.................................................................
        // smlal v12.4S, v13.4H, v14.4H         // .........................*.................................................
        // smlal2 v11.4S, v13.8H, v14.8H        // ..........................*................................................
        // uzp2 v28.8H, v18.8H, v19.8H          // ....*......................................................................
        // smlal v12.4S, v28.4H, v15.4H         // ..............................*............................................
        // smlal2 v11.4S, v28.8H, v15.8H        // ...............................*...........................................
        // ld1 {v22.8H}, [x6], #16              // .............................*.............................................
        // uzp1 v1.8H, v12.8H, v11.8H           // ...................................*.......................................
        // smull v23.4S, v3.4H, v17.4H          // ...................*.......................................................
        // mul v14.8H, v1.8H, v2.8H             // .....................................*.....................................
        // ld1 {v4.8H}, [x3], #16               // ..............*............................................................
        // smlal2 v11.4S, v14.8H, v0.8H         // ........................................*..................................
        // smull2 v20.4S, v3.8H, v17.8H         // ....................*......................................................
        // ldr q18, [x4], #32                   // ..................................*........................................
        // ldr q30, [x5], #32                   // .......................................*...................................
        // smlal2 v20.4S, v10.8H, v4.8H         // ........................*..................................................
        // smlal v12.4S, v14.4H, v0.4H          // .........................................*.................................
        // smlal v23.4S, v10.4H, v4.4H          // .......................*...................................................
        // smlal2 v20.4S, v13.8H, v15.8H        // ............................*..............................................
        // ldr q8, [x2], #32                    // ..........................................*................................
        // smlal v23.4S, v13.4H, v15.4H         // ...........................*...............................................
        // smlal2 v20.4S, v28.8H, v22.8H        // ................................*..........................................
        // ldr q9, [x2, #-16]                   // ...........................................*...............................
        // smlal v23.4S, v28.4H, v22.4H         // .................................*.........................................
        // uzp2 v27.8H, v12.8H, v11.8H          // ..................................................................*........
        // uzp1 v17.8H, v8.8H, v9.8H            // .............................................*.............................
        // uzp2 v4.8H, v8.8H, v9.8H             // ................................................*..........................
        // uzp1 v5.8H, v23.8H, v20.8H           // ....................................*......................................
        // mul v31.8H, v5.8H, v2.8H             // ......................................*....................................
        // ldr q19, [x4, #-16]                  // ............................................*..............................
        // ldr q29, [x1], #32                   // ..............................................*............................
        // ldr q12, [x1, #-16]                  // ...............................................*...........................
        // smlal2 v20.4S, v31.8H, v0.8H         // ..............................................................*............
        // uzp1 v13.8H, v18.8H, v19.8H          // .................................................*.........................
        // uzp1 v3.8H, v29.8H, v12.8H           // ..................................................*........................
        // uzp2 v10.8H, v29.8H, v12.8H          // ...................................................*.......................
        // smull v12.4S, v3.4H, v4.4H           // ....................................................*......................
        // smull2 v11.4S, v3.8H, v4.8H          // .....................................................*.....................
        // ldr q5, [x5, #-16]                   // ......................................................*....................
        // smlal v12.4S, v10.4H, v17.4H         // .......................................................*...................
        // smlal2 v11.4S, v10.8H, v17.8H        // ........................................................*..................
        // uzp2 v14.8H, v30.8H, v5.8H           // .........................................................*.................
        // uzp1 v15.8H, v30.8H, v5.8H           // ..........................................................*................
        // smlal v12.4S, v13.4H, v14.4H         // ...........................................................*...............
        // smlal2 v11.4S, v13.8H, v14.8H        // ............................................................*..............
        // uzp2 v28.8H, v18.8H, v19.8H          // .............................................................*.............
        // smlal v23.4S, v31.4H, v0.4H          // .................................................................*.........
        // smlal v12.4S, v28.4H, v15.4H         // ...............................................................*...........
        // smlal2 v11.4S, v28.8H, v15.8H        // ................................................................*..........
        // ld1 {v22.8H}, [x6], #16              // .......................................................................*...
        // uzp2 v19.8H, v23.8H, v20.8H          // .....................................................................*.....
        // uzp1 v1.8H, v12.8H, v11.8H           // ....................................................................*......
        // smull v23.4S, v3.4H, v17.4H          // ...................................................................*.......
        // mul v14.8H, v1.8H, v2.8H             // ......................................................................*....
        // zip2 v9.8H, v19.8H, v27.8H           // ........................................................................*..
        // ld1 {v4.8H}, [x3], #16               // ..........................................................................*
        // smlal2 v11.4S, v14.8H, v0.8H         // .........................................................................*.

        sub count, count, #2
polyvec_basemul_acc_montgomery_cached_asm_k2_opt_loop:
                                             // Instructions:    48
                                             // Expected cycles: 58
                                             // Expected IPC:    0.83

                                             // Cycle bound:     58.0
                                             // IPC bound:       0.83

                                             // Wall time:     6.39s
                                             // User time:     6.39s

                                             // -------------- original position -------------->
                                             // 0                        25
                                             // |------------------------|----------------------
        smull2 v20.4S, v3.8H, v17.8H         // ..........*.....................................
        ldr q18, [x4], #32                   // .................e..............................
        ldr q30, [x5], #32                   // .....................e..........................
        smlal2 v20.4S, v10.8H, v4.8H         // ............*...................................
        smlal v12.4S, v14.4H, v0.4H          // .........................................*......
        smlal v23.4S, v10.4H, v4.4H          // ...........*....................................
        str q9, [x0, #16]                    // ...............................................l
        smlal2 v20.4S, v13.8H, v15.8H        // ...........................*....................
        ldr q8, [x2], #32                    // ....e...........................................
        smlal v23.4S, v13.4H, v15.4H         // ..........................*.....................
        smlal2 v20.4S, v28.8H, v22.8H        // .............................*..................
        zip1 v26.8H, v19.8H, v27.8H          // ............................................l...
        ldr q9, [x2, #-16]                   // .....e..........................................
        smlal v23.4S, v28.4H, v22.4H         // ............................*...................
        uzp2 v27.8H, v12.8H, v11.8H          // ...........................................*....
        uzp1 v17.8H, v8.8H, v9.8H            // ......e.........................................
        uzp2 v4.8H, v8.8H, v9.8H             // .......e........................................
        uzp1 v5.8H, v23.8H, v20.8H           // ..................................*.............
        str q26, [x0], #32                   // ..............................................l.
        mul v31.8H, v5.8H, v2.8H             // ...................................*............
        ldr q19, [x4, #-16]                  // ..................e.............................
        ldr q29, [x1], #32                   // e...............................................
        ldr q12, [x1, #-16]                  // .e..............................................
        smlal2 v20.4S, v31.8H, v0.8H         // .....................................*..........
        uzp1 v13.8H, v18.8H, v19.8H          // ...................e............................
        uzp1 v3.8H, v29.8H, v12.8H           // ..e.............................................
        uzp2 v10.8H, v29.8H, v12.8H          // ...e............................................
        smull v12.4S, v3.4H, v4.4H           // .............e..................................
        smull2 v11.4S, v3.8H, v4.8H          // ..............e.................................
        ldr q5, [x5, #-16]                   // ......................e.........................
        smlal v12.4S, v10.4H, v17.4H         // ...............e................................
        smlal2 v11.4S, v10.8H, v17.8H        // ................e...............................
        uzp2 v14.8H, v30.8H, v5.8H           // ........................e.......................
        uzp1 v15.8H, v30.8H, v5.8H           // .......................e........................
        smlal v12.4S, v13.4H, v14.4H         // ..............................e.................
        smlal2 v11.4S, v13.8H, v14.8H        // ...............................e................
        uzp2 v28.8H, v18.8H, v19.8H          // ....................e...........................
        smlal v23.4S, v31.4H, v0.4H          // ....................................*...........
        smlal v12.4S, v28.4H, v15.4H         // ................................e...............
        smlal2 v11.4S, v28.8H, v15.8H        // .................................e..............
        ld1 {v22.8H}, [x6], #16              // .........................e......................
        uzp2 v19.8H, v23.8H, v20.8H          // ......................................*.........
        uzp1 v1.8H, v12.8H, v11.8H           // .......................................e........
        smull v23.4S, v3.4H, v17.4H          // .........e......................................
        mul v14.8H, v1.8H, v2.8H             // ........................................e.......
        zip2 v9.8H, v19.8H, v27.8H           // .............................................*..
        ld1 {v4.8H}, [x3], #16               // ........e.......................................
        smlal2 v11.4S, v14.8H, v0.8H         // ..........................................e.....

                                             // ------------------------------------------------- new position -------------------------------------------------->
                                             // 0                        25                       50                       75                       100
                                             // |------------------------|------------------------|------------------------|------------------------|-------------
        // ldr q12, [x1], #32                  // ....................e..........................'....................~..........................'..................
        // ldr q13, [x1, #-16]                 // .....................e.........................'.....................~.........................'..................
        // uzp1 v3.8h, v12.8h, v13.8h          // ........................e......................'........................~......................'..................
        // uzp2 v4.8h, v12.8h, v13.8h          // .........................e.....................'.........................~.....................'..................
        // ldr q12, [x2], #32                  // .......e.......................................'.......~.......................................'.......~..........
        // ldr q13, [x2, #-16]                 // ...........e...................................'...........~...................................'...........~......
        // uzp1 v5.8h, v12.8h, v13.8h          // ..............e................................'..............~................................'..............~...
        // uzp2 v6.8h, v12.8h, v13.8h          // ...............e...............................'...............~...............................'...............~..
        // ld1 {v7.8h}, [x3], #16              // .............................................e.'.............................................~.'..................
        // smull  v8.4s, v3.4h, v5.4h          // ..........................................e....'..........................................~....'..................
        // smull2 v10.4s, v3.8h, v5.8h         // ...............................................*...............................................~..................
        // smlal  v8.4s, v4.4h, v7.4h          // ....~..........................................'....*..........................................'....~.............
        // smlal2 v10.4s, v4.8h, v7.8h         // ..~............................................'..*............................................'..~...............
        // smull  v9.4s, v3.4h, v6.4h          // ..........................e....................'..........................~....................'..................
        // smull2 v11.4s, v3.8h, v6.8h         // ...........................e...................'...........................~...................'..................
        // smlal  v9.4s, v4.4h, v5.4h          // .............................e.................'.............................~.................'..................
        // smlal2 v11.4s, v4.8h, v5.8h         // ..............................e................'..............................~................'..................
        // ldr q12, [x4], #32                  // e..............................................'~..............................................'~.................
        // ldr q13, [x4, #-16]                 // ...................e...........................'...................~...........................'..................
        // uzp1 v3.8h, v12.8h, v13.8h          // .......................e.......................'.......................~.......................'..................
        // uzp2 v4.8h, v12.8h, v13.8h          // ...................................e...........'...................................~...........'..................
        // ldr q12, [x5], #32                  // .e.............................................'.~.............................................'.~................
        // ldr q13, [x5, #-16]                 // ............................e..................'............................~..................'..................
        // uzp1 v5.8h, v12.8h, v13.8h          // ................................e..............'................................~..............'..................
        // uzp2 v6.8h, v12.8h, v13.8h          // ...............................e...............'...............................~...............'..................
        // ld1 {v7.8h}, [x6], #16              // .......................................e.......'.......................................~.......'..................
        // smlal  v8.4s, v3.4h, v5.4h          // ........~......................................'........*......................................'........~.........
        // smlal2 v10.4s, v3.8h, v5.8h         // ......~........................................'......*........................................'......~...........
        // smlal  v8.4s, v4.4h, v7.4h          // ............~..................................'............*..................................'............~.....
        // smlal2 v10.4s, v4.8h, v7.8h         // .........~.....................................'.........*.....................................'.........~........
        // smlal  v9.4s, v3.4h, v6.4h          // .................................e.............'.................................~.............'..................
        // smlal2 v11.4s, v3.8h, v6.8h         // ..................................e............'..................................~............'..................
        // smlal  v9.4s, v4.4h, v5.4h          // .....................................e.........'.....................................~.........'..................
        // smlal2 v11.4s, v4.8h, v5.8h         // ......................................e........'......................................~........'..................
        // uzp1   v28.8h, v8.8h, v10.8h        // ................~..............................'................*..............................'................~.
        // mul    v28.8h, v28.8h, v2.8h        // ..................~............................'..................*............................'..................
        // smlal  v8.4s, v28.4h, v0.4h         // ....................................~..........'....................................*..........'..................
        // smlal2 v10.4s, v28.8h, v0.8h        // ......................~........................'......................*........................'..................
        // uzp2   v26.8h, v8.8h, v10.8h        // ........................................~......'........................................*......'..................
        // uzp1   v28.8h, v9.8h, v11.8h        // .........................................e.....'.........................................~.....'..................
        // mul    v28.8h, v28.8h, v2.8h        // ...........................................e...'...........................................~...'..................
        // smlal  v9.4s, v28.4h, v0.4h         // ...~...........................................'...*...........................................'...~..............
        // smlal2 v11.4s, v28.8h, v0.8h        // ..............................................e'..............................................~'..................
        // uzp2   v27.8h, v9.8h, v11.8h        // .............~.................................'.............*.................................'.............~....
        // zip1 v12.8h, v26.8h, v27.8h         // ..........~....................................'..........~....................................'..........l.......
        // zip2 v13.8h, v26.8h, v27.8h         // ............................................~..'............................................*..'..................
        // str q12, [x0], #32                  // .................~.............................'.................~.............................'.................l
        // str q13, [x0, #-16]                 // .....~.........................................'.....~.........................................'.....l............

        sub count, count, #1
        cbnz count, polyvec_basemul_acc_montgomery_cached_asm_k2_opt_loop
                                            // Instructions:    21
                                            // Expected cycles: 35
                                            // Expected IPC:    0.60

                                            // Cycle bound:     35.0
                                            // IPC bound:       0.60

                                            // Wall time:     0.08s
                                            // User time:     0.08s

                                            // ----- original position ----->
                                            // 0                        25
                                            // |------------------------|----
        smull2 v5.4S, v3.8H, v17.8H         // *.............................
        smlal v12.4S, v14.4H, v0.4H         // ..*...........................
        smlal v23.4S, v10.4H, v4.4H         // ...*..........................
        str q9, [x0, #16]                   // ....*.........................
        smlal2 v5.4S, v10.8H, v4.8H         // .*............................
        uzp2 v11.8H, v12.8H, v11.8H         // ..........*...................
        zip1 v9.8H, v19.8H, v27.8H          // ........*.....................
        smlal v23.4S, v13.4H, v15.4H        // ......*.......................
        smlal2 v5.4S, v13.8H, v15.8H        // .....*........................
        str q9, [x0], #32                   // ............*.................
        smlal v23.4S, v28.4H, v22.4H        // .........*....................
        smlal2 v5.4S, v28.8H, v22.8H        // .......*......................
        uzp1 v9.8H, v23.8H, v5.8H           // ...........*..................
        mul v9.8H, v9.8H, v2.8H             // .............*................
        smlal2 v5.4S, v9.8H, v0.8H          // ..............*...............
        smlal v23.4S, v9.4H, v0.4H          // ...............*..............
        uzp2 v9.8H, v23.8H, v5.8H           // ................*.............
        zip2 v5.8H, v9.8H, v11.8H           // .................*............
        zip1 v9.8H, v9.8H, v11.8H           // ...................*..........
        str q5, [x0, #16]                   // ..................*...........
        str q9, [x0], #32                   // ....................*.........

                                              // -------- new position -------->
                                              // 0                        25
                                              // |------------------------|-----
        // smull2 v20.4S, v3.8H, v17.8H         // *..............................
        // smlal2 v20.4S, v10.8H, v4.8H         // ....*..........................
        // smlal v12.4S, v14.4H, v0.4H          // .*.............................
        // smlal v23.4S, v10.4H, v4.4H          // ..*............................
        // str q9, [x0, #16]                    // ...*...........................
        // smlal2 v20.4S, v13.8H, v15.8H        // ........*......................
        // smlal v23.4S, v13.4H, v15.4H         // .......*.......................
        // smlal2 v20.4S, v28.8H, v22.8H        // ...........*...................
        // zip1 v26.8H, v19.8H, v27.8H          // ......*........................
        // smlal v23.4S, v28.4H, v22.4H         // ..........*....................
        // uzp2 v27.8H, v12.8H, v11.8H          // .....*.........................
        // uzp1 v5.8H, v23.8H, v20.8H           // ............*..................
        // str q26, [x0], #32                   // .........*.....................
        // mul v31.8H, v5.8H, v2.8H             // .............*.................
        // smlal2 v20.4S, v31.8H, v0.8H         // ..............*................
        // smlal v23.4S, v31.4H, v0.4H          // ...............*...............
        // uzp2 v19.8H, v23.8H, v20.8H          // ................*..............
        // zip2 v9.8H, v19.8H, v27.8H           // .................*.............
        // str q9, [x0, #16]                    // ...................*...........
        // zip1 v26.8H, v19.8H, v27.8H          // ..................*............
        // str q26, [x0], #32                   // ....................*..........


        pop_stack
        ret

/****************** REGISTER DEALLOCATIONS *******************/
    .unreq out
    .unreq a0_ptr
    .unreq b0_ptr
    .unreq b0_cache_ptr
    .unreq a1_ptr
    .unreq b1_ptr
    .unreq b1_cache_ptr
    .unreq a2_ptr
    .unreq b2_ptr
    .unreq b2_cache_ptr
    .unreq a3_ptr
    .unreq b3_ptr
    .unreq b3_cache_ptr
    .unreq count
    .unreq modulus
    .unreq modulus_twisted
    .unreq wtmp
    .unreq aa0
    .unreq aa1
    .unreq bb0
    .unreq bb1
    .unreq bb1t
    .unreq res0l
    .unreq res1l
    .unreq res0h
    .unreq res1h
    .unreq tmp0
    .unreq tmp1
    .unreq q_tmp0
    .unreq q_tmp1
    .unreq out0
    .unreq out1
    .unreq t0

/* simpasm: footer-start */
#endif /* defined(MLKEM_NATIVE_ARITH_BACKEND_AARCH64_OPT) && \
          !defined(MLKEM_NATIVE_MULTILEVEL_BUILD_NO_SHARED) && \
          (defined(MLKEM_NATIVE_MULTILEVEL_BUILD_WITH_SHARED) || MLKEM_K == 2) */
