/*
 * Copyright (c) 2024-2025 The mlkem-native project authors
 * SPDX-License-Identifier: Apache-2.0
 */

// AArch64 re-implementation of the asymmetric base multiplication from:

// Neon NTT: Faster Dilithium, Kyber, and Saber on Cortex-A72 and Apple M1
// https://eprint.iacr.org/2021/986
// https://github.com/neon-ntt/neon-ntt

#include "../../../common.h"
#if defined(MLKEM_NATIVE_ARITH_BACKEND_AARCH64_OPT) && \
    !defined(MLKEM_NATIVE_MULTILEVEL_BUILD_NO_SHARED) && \
    (defined(MLKEM_NATIVE_MULTILEVEL_BUILD_WITH_SHARED) || MLKEM_K == 4)
/* simpasm: header-end */

// Input:
// - Vectors al, ah of 32-bit entries
// Output:
// - Montgomery reductions of al || ah, stored in al
.macro montgomery_reduce_long x, a
        uzp1   t0.8h, \a\()l.8h, \a\()h.8h
        mul    t0.8h, t0.8h, modulus_twisted.8h
        smlal  \a\()l.4s, t0.4h, modulus.4h
        smlal2 \a\()h.4s, t0.8h, modulus.8h
        uzp2   \x\().8h, \a\()l.8h, \a\()h.8h
.endm

// Computes products (a0*b0 + a0*b0t, a0*b1 + a1*b0) in 32-bit.

// Bounds:
// - Assume |a| < 4096,
// - Result: < 2*4096*2^15 = 2^28
.macro pmull d, a, b
        smull  \d\()0l.4s, \a\()0.4h, \b\()0.4h
        smull2 \d\()0h.4s, \a\()0.8h, \b\()0.8h
        smlal  \d\()0l.4s, \a\()1.4h, \b\()1t.4h
        smlal2 \d\()0h.4s, \a\()1.8h, \b\()1t.8h

        smull  \d\()1l.4s, \a\()0.4h, \b\()1.4h
        smull2 \d\()1h.4s, \a\()0.8h, \b\()1.8h
        smlal  \d\()1l.4s, \a\()1.4h, \b\()0.4h
        smlal2 \d\()1h.4s, \a\()1.8h, \b\()0.8h
.endm

.macro pmlal d, a, b
        smlal  \d\()0l.4s, \a\()0.4h, \b\()0.4h
        smlal2 \d\()0h.4s, \a\()0.8h, \b\()0.8h
        smlal  \d\()0l.4s, \a\()1.4h, \b\()1t.4h
        smlal2 \d\()0h.4s, \a\()1.8h, \b\()1t.8h

        smlal  \d\()1l.4s, \a\()0.4h, \b\()1.4h
        smlal2 \d\()1h.4s, \a\()0.8h, \b\()1.8h
        smlal  \d\()1l.4s, \a\()1.4h, \b\()0.4h
        smlal2 \d\()1h.4s, \a\()1.8h, \b\()0.8h
.endm

.macro ld2_wrap a, ptr
        ldr q_tmp0, [\ptr\()], #32
        ldr q_tmp1, [\ptr\(), #-16]
        uzp1 \a\()0.8h, tmp0.8h, tmp1.8h
        uzp2 \a\()1.8h, tmp0.8h, tmp1.8h
.endm

.macro st2_wrap a, ptr
        zip1 tmp0.8h, \a\()0.8h, \a\()1.8h
        zip2 tmp1.8h, \a\()0.8h, \a\()1.8h
        str q_tmp0, [\ptr\()], #32
        str q_tmp1, [\ptr\(), #-16]
.endm

.macro load_polys a, b, a_ptr, b_ptr, b_cache_ptr
        ld2_wrap \a\(), \a_ptr
        ld2_wrap \b\(), \b_ptr
        ld1 {\b\()1t.8h}, [\b_cache_ptr], #16
.endm

.macro save_vregs
        sub sp, sp, #(16*4)
        stp  d8,  d9, [sp, #16*0]
        stp d10, d11, [sp, #16*1]
        stp d12, d13, [sp, #16*2]
        stp d14, d15, [sp, #16*3]
.endm

.macro restore_vregs
        ldp  d8,  d9, [sp, #16*0]
        ldp d10, d11, [sp, #16*1]
        ldp d12, d13, [sp, #16*2]
        ldp d14, d15, [sp, #16*3]
        add sp, sp, #(16*4)
.endm

.macro push_stack
        save_vregs
.endm

.macro pop_stack
        restore_vregs
.endm

        out          .req x0
        a0_ptr       .req x1
        b0_ptr       .req x2
        b0_cache_ptr .req x3
        a1_ptr       .req x4
        b1_ptr       .req x5
        b1_cache_ptr .req x6
        a2_ptr       .req x7
        b2_ptr       .req x8
        b2_cache_ptr .req x9
        a3_ptr       .req x10
        b3_ptr       .req x11
        b3_cache_ptr .req x12
        count        .req x13
        wtmp         .req w14

        modulus           .req v0
        modulus_twisted   .req v2

        aa0      .req v3
        aa1      .req v4
        bb0      .req v5
        bb1      .req v6
        bb1t     .req v7

        res0l   .req v8
        res1l   .req v9
        res0h   .req v10
        res1h   .req v11

        tmp0 .req v12
        tmp1 .req v13
        q_tmp0 .req q12
        q_tmp1 .req q13

        out0 .req v26
        out1 .req v27

        t0   .req v28

        .text
        .global MLKEM_ASM_NAMESPACE(polyvec_basemul_acc_montgomery_cached_asm_k4_opt)
        .balign 4
MLKEM_ASM_NAMESPACE(polyvec_basemul_acc_montgomery_cached_asm_k4_opt):
        push_stack
        mov wtmp, #3329
        dup modulus.8h, wtmp

        mov wtmp, #3327
        dup modulus_twisted.8h, wtmp

        // Computed bases of vector entries

        add a1_ptr, a0_ptr, #(1 * 512)
        add b1_ptr, b0_ptr, #(1 * 512)
        add b1_cache_ptr, b0_cache_ptr, #(1 * 512/2)
        add a2_ptr, a0_ptr, #(2 * 512)
        add b2_ptr, b0_ptr, #(2 * 512)
        add b2_cache_ptr, b0_cache_ptr, #(2 * 512/2)
        add a3_ptr, a0_ptr, #(3 * 512)
        add b3_ptr, b0_ptr, #(3 * 512)
        add b3_cache_ptr, b0_cache_ptr, #(3 * 512/2)

        // Bounds:

        // Each pmull is bound by 2*4096*2^15=2^28, so the final value
        // before Montgomery reduction is bound by 2^30.

        mov count, #(MLKEM_N / 16)
                                             // Instructions:    114
                                             // Expected cycles: 153
                                             // Expected IPC:    0.75
                                             //
                                             // Cycle bound:     153.0
                                             // IPC bound:       0.75
                                             //
                                             // Wall time:     0.69s
                                             // User time:     0.69s
                                             //
                                             // ----------------------------------------------- original position ----------------------------------------------->
                                             // 0                        25                       50                       75                       100
                                             // |------------------------|------------------------|------------------------|------------------------|-------------
        ldr q23, [x2, #16]                   // .*................................................................................................................
        ldr q19, [x2], #32                   // *.................................................................................................................
        ldr q17, [x5], #32                   // ..*...............................................................................................................
        uzp2 v13.8H, v19.8H, v23.8H          // ..........*.......................................................................................................
        uzp1 v19.8H, v19.8H, v23.8H          // ...........*......................................................................................................
        ldr q23, [x5, #-16]                  // ...*..............................................................................................................
        ldr q30, [x1, #16]                   // .....*............................................................................................................
        uzp2 v9.8H, v17.8H, v23.8H           // ....*.............................................................................................................
        uzp1 v23.8H, v17.8H, v23.8H          // .......*..........................................................................................................
        ldr q17, [x1], #32                   // ......*...........................................................................................................
        ldr q10, [x7, #16]                   // .............*....................................................................................................
        uzp1 v12.8H, v17.8H, v30.8H          // ........*.........................................................................................................
        uzp2 v17.8H, v17.8H, v30.8H          // .........*........................................................................................................
        smull2 v30.4S, v12.8H, v13.8H        // ............*.....................................................................................................
        smull v13.4S, v12.4H, v13.4H         // ............................................*.....................................................................
        smull2 v22.4S, v12.8H, v19.8H        // .....................................*............................................................................
        smull v12.4S, v12.4H, v19.4H         // ..........................................*.......................................................................
        smlal2 v30.4S, v17.8H, v19.8H        // ...............................*..................................................................................
        smlal v13.4S, v17.4H, v19.4H         // ...............................................*..................................................................
        ldr q19, [x4], #32                   // ....................*.............................................................................................
        ldr q16, [x4, #-16]                  // .....................*............................................................................................
        ld1 {v8.8H}, [x3], #16               // ................................*.................................................................................
        uzp1 v26.8H, v19.8H, v16.8H          // .......................*..........................................................................................
        uzp2 v19.8H, v19.8H, v16.8H          // ........................*.........................................................................................
        smlal2 v30.4S, v26.8H, v9.8H         // .................................*................................................................................
        smlal v13.4S, v26.4H, v9.4H          // ..................................................*...............................................................
        smlal2 v22.4S, v17.8H, v8.8H         // ........................................*.........................................................................
        smlal v12.4S, v17.4H, v8.4H          // .................................................*................................................................
        smlal2 v30.4S, v19.8H, v23.8H        // ...................................*..............................................................................
        smlal v13.4S, v19.4H, v23.4H         // .......................................................*..........................................................
        smlal2 v22.4S, v26.8H, v23.8H        // ...........................................*......................................................................
        smlal v12.4S, v26.4H, v23.4H         // .....................................................*............................................................
        ldr q23, [x7], #32                   // ......................*...........................................................................................
        ldr q17, [x8, #16]                   // ..............*...................................................................................................
        uzp1 v9.8H, v23.8H, v10.8H           // ..........................*.......................................................................................
        uzp2 v23.8H, v23.8H, v10.8H          // ....................................*.............................................................................
        ldr q10, [x10], #32                  // ...............*..................................................................................................
        ldr q16, [x10, #-16]                 // ................*.................................................................................................
        ld1 {v8.8H}, [x12], #16              // .................*................................................................................................
        uzp1 v26.8H, v10.8H, v16.8H          // ..................*...............................................................................................
        uzp2 v10.8H, v10.8H, v16.8H          // ...................*..............................................................................................
        ld1 {v16.8H}, [x6], #16              // .........................*........................................................................................
        ldr q3, [x11, #16]                   // ...........................*......................................................................................
        smlal2 v22.4S, v19.8H, v16.8H        // ..............................................*...................................................................
        smlal v12.4S, v19.4H, v16.4H         // ........................................................*.........................................................
        ldr q19, [x11], #32                  // ............................*.....................................................................................
        ld1 {v16.8H}, [x9], #16              // .............................*....................................................................................
        uzp1 v4.8H, v19.8H, v3.8H            // ..................................*...............................................................................
        uzp2 v19.8H, v19.8H, v3.8H           // .......................................*..........................................................................
        ldr q3, [x8], #32                    // ..............................*...................................................................................
        ldr q31, [x2], #32                   // ......................................*...........................................................................
        uzp1 v6.8H, v3.8H, v17.8H            // ...................................................*..............................................................
        uzp2 v17.8H, v3.8H, v17.8H           // .........................................................*........................................................
        smlal2 v22.4S, v9.8H, v6.8H          // ..........................................................*.......................................................
        smlal2 v30.4S, v9.8H, v17.8H         // ...........................................................*......................................................
        smlal v13.4S, v9.4H, v17.4H          // ............................................................*.....................................................
        smlal v12.4S, v9.4H, v6.4H           // .............................................................*....................................................
        smlal2 v22.4S, v23.8H, v16.8H        // ..............................................................*...................................................
        smlal2 v30.4S, v23.8H, v6.8H         // ...............................................................*..................................................
        smlal v13.4S, v23.4H, v6.4H          // ................................................................*.................................................
        smlal v12.4S, v23.4H, v16.4H         // .................................................................*................................................
        smlal2 v22.4S, v26.8H, v4.8H         // ..................................................................*...............................................
        smlal2 v30.4S, v26.8H, v19.8H        // ...................................................................*..............................................
        smlal v13.4S, v26.4H, v19.4H         // ....................................................................*.............................................
        smlal v12.4S, v26.4H, v4.4H          // .....................................................................*............................................
        smlal2 v22.4S, v10.8H, v8.8H         // ......................................................................*...........................................
        smlal2 v30.4S, v10.8H, v4.8H         // .......................................................................*..........................................
        smlal v13.4S, v10.4H, v4.4H          // ........................................................................*.........................................
        smlal v12.4S, v10.4H, v8.4H          // .........................................................................*........................................
        ldr q19, [x2, #-16]                  // .........................................*........................................................................
        uzp1 v23.8H, v13.8H, v30.8H          // ...........................................................................*......................................
        uzp1 v17.8H, v12.8H, v22.8H          // ....................................................................................*.............................
        mul v23.8H, v23.8H, v2.8H            // .............................................................................*....................................
        uzp2 v21.8H, v31.8H, v19.8H          // ................................................................................*.................................
        uzp1 v19.8H, v31.8H, v19.8H          // ...................................................................................*..............................
        mul v17.8H, v17.8H, v2.8H            // .....................................................................................*............................
        smlal v13.4S, v23.4H, v0.4H          // .................................................................................*................................
        smlal2 v30.4S, v23.8H, v0.8H         // ..................................................................................*...............................
        ldr q23, [x5], #32                   // .............................................*....................................................................
        smlal2 v22.4S, v17.8H, v0.8H         // ...........................................................................................................*......
        uzp2 v15.8H, v13.8H, v30.8H          // ......................................................................................*...........................
        smlal v12.4S, v17.4H, v0.4H          // ............................................................................................................*.....
        ldr q17, [x5, #-16]                  // ................................................*.................................................................
        ldr q13, [x1, #16]                   // ......................................................*...........................................................
        uzp2 v27.8H, v23.8H, v17.8H          // ....................................................*.............................................................
        uzp1 v28.8H, v23.8H, v17.8H          // ............................................................................*.....................................
        uzp2 v7.8H, v12.8H, v22.8H           // ...............................................................................................................*..
        ldr q23, [x1], #32                   // ..........................................................................*.......................................
        zip1 v5.8H, v7.8H, v15.8H            // .................................................................................................................*
        ldr q3, [x7, #16]                    // ........................................................................................*.........................
        uzp1 v31.8H, v23.8H, v13.8H          // ..............................................................................*...................................
        uzp2 v16.8H, v23.8H, v13.8H          // ...............................................................................*..................................
        smull2 v24.4S, v31.8H, v21.8H        // .......................................................................................*..........................
        ldr q6, [x8, #16]                    // .........................................................................................*........................
        ldr q23, [x10], #32                  // ..........................................................................................*.......................
        smlal2 v24.4S, v16.8H, v19.8H        // ..........................................................................................................*.......
        ldr q17, [x10, #-16]                 // ...........................................................................................*......................
        ld1 {v22.8H}, [x12], #16             // ............................................................................................*.....................
        uzp1 v30.8H, v23.8H, v17.8H          // .............................................................................................*....................
        uzp2 v11.8H, v23.8H, v17.8H          // ..............................................................................................*...................
        ldr q23, [x4], #32                   // ...............................................................................................*..................
        ldr q17, [x4, #-16]                  // ................................................................................................*.................
        ldr q4, [x7], #32                    // .................................................................................................*................
        uzp1 v20.8H, v23.8H, v17.8H          // ..................................................................................................*...............
        uzp2 v26.8H, v23.8H, v17.8H          // ...................................................................................................*..............
        uzp1 v9.8H, v4.8H, v3.8H             // .....................................................................................................*............
        smlal2 v24.4S, v20.8H, v27.8H        // ..............................................................................................................*...
        ld1 {v8.8H}, [x6], #16               // ....................................................................................................*.............
        ldr q25, [x11, #16]                  // ......................................................................................................*...........
        ldr q29, [x11], #32                  // .......................................................................................................*..........
        ld1 {v12.8H}, [x9], #16              // ........................................................................................................*.........
        uzp1 v10.8H, v29.8H, v25.8H          // ................................................................................................................*.
        ldr q14, [x8], #32                   // .........................................................................................................*........
        ld1 {v23.8H}, [x3], #16              // .............................................................................................................*....

                                              // ------------------------------------------------- new position -------------------------------------------------->
                                              // 0                        25                       50                       75                       100
                                              // |------------------------|------------------------|------------------------|------------------------|-------------
        // ldr q3, [x2], #32                  // .*................................................................................................................
        // ldr q17, [x2, #-16]                // *.................................................................................................................
        // ldr q21, [x5], #32                 // ..*...............................................................................................................
        // ldr q19, [x5, #-16]                // .....*............................................................................................................
        // uzp2 v27.8H, v21.8H, v19.8H        // .......*..........................................................................................................
        // ldr q25, [x1, #16]                 // ......*...........................................................................................................
        // ldr q22, [x1], #32                 // .........*........................................................................................................
        // uzp1 v28.8H, v21.8H, v19.8H        // ........*.........................................................................................................
        // uzp1 v31.8H, v22.8H, v25.8H        // ...........*......................................................................................................
        // uzp2 v16.8H, v22.8H, v25.8H        // ............*.....................................................................................................
        // uzp2 v21.8H, v3.8H, v17.8H         // ...*..............................................................................................................
        // uzp1 v19.8H, v3.8H, v17.8H         // ....*.............................................................................................................
        // smull2 v24.4S, v31.8H, v21.8H      // .............*....................................................................................................
        // ldr q3, [x7, #16]                  // ..........*.......................................................................................................
        // ldr q6, [x8, #16]                  // .................................*................................................................................
        // ldr q8, [x10], #32                 // ....................................*.............................................................................
        // ldr q26, [x10, #-16]               // .....................................*............................................................................
        // ld1 {v22.8H}, [x12], #16           // ......................................*...........................................................................
        // uzp1 v30.8H, v8.8H, v26.8H         // .......................................*..........................................................................
        // uzp2 v11.8H, v8.8H, v26.8H         // ........................................*.........................................................................
        // ldr q8, [x4], #32                  // ...................*..............................................................................................
        // ldr q26, [x4, #-16]                // ....................*.............................................................................................
        // ldr q4, [x7], #32                  // ................................*.................................................................................
        // uzp1 v20.8H, v8.8H, v26.8H         // ......................*...........................................................................................
        // uzp2 v26.8H, v8.8H, v26.8H         // .......................*..........................................................................................
        // ld1 {v8.8H}, [x6], #16             // .........................................*........................................................................
        // uzp1 v9.8H, v4.8H, v3.8H           // ..................................*...............................................................................
        // ldr q25, [x11, #16]                // ..........................................*.......................................................................
        // ldr q29, [x11], #32                // .............................................*....................................................................
        // ld1 {v12.8H}, [x9], #16            // ..............................................*...................................................................
        // ldr q14, [x8], #32                 // .................................................*................................................................
        // smlal2 v24.4S, v16.8H, v19.8H      // .................*................................................................................................
        // ld1 {v23.8H}, [x3], #16            // .....................*............................................................................................
        // smlal2 v24.4S, v20.8H, v27.8H      // ........................*.........................................................................................
        // uzp1 v10.8H, v29.8H, v25.8H        // ...............................................*..................................................................
        // smlal2 v24.4S, v26.8H, v28.8H      // ............................*.....................................................................................
        // uzp2 v4.8H, v4.8H, v3.8H           // ...................................*..............................................................................
        // smull2 v13.4S, v31.8H, v19.8H      // ...............*..................................................................................................
        // ldr q3, [x2], #32                  // ..................................................*...............................................................
        // uzp2 v1.8H, v29.8H, v25.8H         // ................................................*.................................................................
        // smlal2 v13.4S, v16.8H, v23.8H      // ..........................*.......................................................................................
        // ldr q17, [x2, #-16]                // .....................................................................*............................................
        // smull v18.4S, v31.4H, v19.4H       // ................*.................................................................................................
        // smlal2 v13.4S, v20.8H, v28.8H      // ..............................*...................................................................................
        // smull v29.4S, v31.4H, v21.4H       // ..............*...................................................................................................
        // ldr q21, [x5], #32                 // ..............................................................................*...................................
        // smlal2 v13.4S, v26.8H, v8.8H       // ...........................................*......................................................................
        // smlal v29.4S, v16.4H, v19.4H       // ..................*...............................................................................................
        // ldr q19, [x5, #-16]                // ..................................................................................*...............................
        // smlal v18.4S, v16.4H, v23.4H       // ...........................*......................................................................................
        // smlal v29.4S, v20.4H, v27.4H       // .........................*........................................................................................
        // uzp1 v31.8H, v14.8H, v6.8H         // ...................................................*..............................................................
        // uzp2 v27.8H, v21.8H, v19.8H        // ....................................................................................*.............................
        // smlal v18.4S, v20.4H, v28.4H       // ...............................*..................................................................................
        // ldr q25, [x1, #16]                 // ...................................................................................*..............................
        // smlal v29.4S, v26.4H, v28.4H       // .............................*....................................................................................
        // smlal v18.4S, v26.4H, v8.4H        // ............................................*.....................................................................
        // uzp2 v26.8H, v14.8H, v6.8H         // ....................................................*.............................................................
        // smlal2 v13.4S, v9.8H, v31.8H       // .....................................................*............................................................
        // smlal2 v24.4S, v9.8H, v26.8H       // ......................................................*...........................................................
        // smlal v29.4S, v9.4H, v26.4H        // .......................................................*..........................................................
        // smlal v18.4S, v9.4H, v31.4H        // ........................................................*.........................................................
        // smlal2 v13.4S, v4.8H, v12.8H       // .........................................................*........................................................
        // smlal2 v24.4S, v4.8H, v31.8H       // ..........................................................*.......................................................
        // smlal v29.4S, v4.4H, v31.4H        // ...........................................................*......................................................
        // smlal v18.4S, v4.4H, v12.4H        // ............................................................*.....................................................
        // smlal2 v13.4S, v30.8H, v10.8H      // .............................................................*....................................................
        // smlal2 v24.4S, v30.8H, v1.8H       // ..............................................................*...................................................
        // smlal v29.4S, v30.4H, v1.4H        // ...............................................................*..................................................
        // smlal v18.4S, v30.4H, v10.4H       // ................................................................*.................................................
        // smlal2 v13.4S, v11.8H, v22.8H      // .................................................................*................................................
        // smlal2 v24.4S, v11.8H, v10.8H      // ..................................................................*...............................................
        // smlal v29.4S, v11.4H, v10.4H       // ...................................................................*..............................................
        // smlal v18.4S, v11.4H, v22.4H       // ....................................................................*.............................................
        // ldr q22, [x1], #32                 // .......................................................................................*..........................
        // uzp1 v31.8H, v29.8H, v24.8H        // ......................................................................*...........................................
        // uzp1 v28.8H, v21.8H, v19.8H        // .....................................................................................*............................
        // mul v19.8H, v31.8H, v2.8H          // ........................................................................*.........................................
        // uzp1 v31.8H, v22.8H, v25.8H        // ..........................................................................................*.......................
        // uzp2 v16.8H, v22.8H, v25.8H        // ...........................................................................................*......................
        // uzp2 v21.8H, v3.8H, v17.8H         // .........................................................................*........................................
        // smlal v29.4S, v19.4H, v0.4H        // ............................................................................*.....................................
        // smlal2 v24.4S, v19.8H, v0.8H       // .............................................................................*....................................
        // uzp1 v19.8H, v3.8H, v17.8H         // ..........................................................................*.......................................
        // uzp1 v26.8H, v18.8H, v13.8H        // .......................................................................*..........................................
        // mul v23.8H, v26.8H, v2.8H          // ...........................................................................*......................................
        // uzp2 v15.8H, v29.8H, v24.8H        // ................................................................................*.................................
        // smull2 v24.4S, v31.8H, v21.8H      // ............................................................................................*.....................
        // ldr q3, [x7, #16]                  // .........................................................................................*........................
        // ldr q6, [x8, #16]                  // .............................................................................................*....................
        // ldr q8, [x10], #32                 // ..............................................................................................*...................
        // ldr q26, [x10, #-16]               // ................................................................................................*.................
        // ld1 {v22.8H}, [x12], #16           // .................................................................................................*................
        // uzp1 v30.8H, v8.8H, v26.8H         // ..................................................................................................*...............
        // uzp2 v11.8H, v8.8H, v26.8H         // ...................................................................................................*..............
        // ldr q8, [x4], #32                  // ....................................................................................................*.............
        // ldr q26, [x4, #-16]                // .....................................................................................................*............
        // ldr q4, [x7], #32                  // ......................................................................................................*...........
        // uzp1 v20.8H, v8.8H, v26.8H         // .......................................................................................................*..........
        // uzp2 v26.8H, v8.8H, v26.8H         // ........................................................................................................*.........
        // ld1 {v8.8H}, [x6], #16             // ...........................................................................................................*......
        // uzp1 v9.8H, v4.8H, v3.8H           // .........................................................................................................*........
        // ldr q25, [x11, #16]                // ............................................................................................................*.....
        // ldr q29, [x11], #32                // .............................................................................................................*....
        // ld1 {v12.8H}, [x9], #16            // ..............................................................................................................*...
        // ldr q14, [x8], #32                 // ................................................................................................................*.
        // smlal2 v24.4S, v16.8H, v19.8H      // ...............................................................................................*..................
        // smlal2 v13.4S, v23.8H, v0.8H       // ...............................................................................*..................................
        // smlal v18.4S, v23.4H, v0.4H        // .................................................................................*................................
        // ld1 {v23.8H}, [x3], #16            // .................................................................................................................*
        // smlal2 v24.4S, v20.8H, v27.8H      // ..........................................................................................................*.......
        // uzp2 v7.8H, v18.8H, v13.8H         // ......................................................................................*...........................
        // uzp1 v10.8H, v29.8H, v25.8H        // ...............................................................................................................*..
        // zip1 v5.8H, v7.8H, v15.8H          // ........................................................................................*.........................

        sub count, count, #2
polyvec_basemul_acc_montgomery_cached_asm_k4_opt_loop:
                                             // Instructions:    82
                                             // Expected cycles: 102
                                             // Expected IPC:    0.80
                                             //
                                             // Cycle bound:     102.0
                                             // IPC bound:       0.80
                                             //
                                             // Wall time:     15.93s
                                             // User time:     15.93s
                                             //
                                             // ------------------------------- original position ------------------------------->
                                             // 0                        25                       50                       75
                                             // |------------------------|------------------------|------------------------|------
        smlal2 v24.4S, v26.8H, v28.8H        // .................................*................................................
        uzp2 v4.8H, v4.8H, v3.8H             // .....................................*............................................
        smull2 v13.4S, v31.8H, v19.8H        // ..........*.......................................................................
        ldr q3, [x2], #32                    // ....e.............................................................................
        uzp2 v1.8H, v29.8H, v25.8H           // ..........................................................*.......................
        smlal2 v13.4S, v16.8H, v23.8H        // ............*.....................................................................
        ldr q17, [x2, #-16]                  // .....e............................................................................
        smull v18.4S, v31.4H, v19.4H         // .........*........................................................................
        smlal2 v13.4S, v20.8H, v28.8H        // ...........................*......................................................
        smull v29.4S, v31.4H, v21.4H         // .............*....................................................................
        ldr q21, [x5], #32                   // .....................e............................................................
        smlal2 v13.4S, v26.8H, v8.8H         // .............................*....................................................
        smlal v29.4S, v16.4H, v19.4H         // ...............*..................................................................
        ldr q19, [x5, #-16]                  // ......................e...........................................................
        smlal v18.4S, v16.4H, v23.4H         // ...........*......................................................................
        smlal v29.4S, v20.4H, v27.4H         // ..............................*...................................................
        uzp1 v31.8H, v14.8H, v6.8H           // ........................................*.........................................
        uzp2 v27.8H, v21.8H, v19.8H          // ........................e.........................................................
        smlal v18.4S, v20.4H, v28.4H         // ..........................*.......................................................
        ldr q25, [x1, #16]                   // .e................................................................................
        smlal v29.4S, v26.4H, v28.4H         // ................................*.................................................
        smlal v18.4S, v26.4H, v8.4H          // ............................*.....................................................
        uzp2 v26.8H, v14.8H, v6.8H           // .........................................*........................................
        smlal2 v13.4S, v9.8H, v31.8H         // ............................................*.....................................
        smlal2 v24.4S, v9.8H, v26.8H         // ................................................*.................................
        smlal v29.4S, v9.4H, v26.4H          // ...............................................*..................................
        smlal v18.4S, v9.4H, v31.4H          // ...........................................*......................................
        smlal2 v13.4S, v4.8H, v12.8H         // ..............................................*...................................
        smlal2 v24.4S, v4.8H, v31.8H         // ..................................................*...............................
        smlal v29.4S, v4.4H, v31.4H          // .................................................*................................
        smlal v18.4S, v4.4H, v12.4H          // .............................................*....................................
        smlal2 v13.4S, v30.8H, v10.8H        // .............................................................*....................
        smlal2 v24.4S, v30.8H, v1.8H         // .................................................................*................
        smlal v29.4S, v30.4H, v1.4H          // ................................................................*.................
        smlal v18.4S, v30.4H, v10.4H         // ............................................................*.....................
        smlal2 v13.4S, v11.8H, v22.8H        // ...............................................................*..................
        smlal2 v24.4S, v11.8H, v10.8H        // ...................................................................*..............
        smlal v29.4S, v11.4H, v10.4H         // ..................................................................*...............
        smlal v18.4S, v11.4H, v22.4H         // ..............................................................*...................
        ldr q22, [x1], #32                   // e.................................................................................
        uzp1 v31.8H, v29.8H, v24.8H          // .........................................................................*........
        uzp1 v28.8H, v21.8H, v19.8H          // .......................e..........................................................
        mul v19.8H, v31.8H, v2.8H            // ..........................................................................*.......
        uzp1 v31.8H, v22.8H, v25.8H          // ..e...............................................................................
        uzp2 v16.8H, v22.8H, v25.8H          // ...e..............................................................................
        uzp2 v21.8H, v3.8H, v17.8H           // .......e..........................................................................
        smlal v29.4S, v19.4H, v0.4H          // ...........................................................................*......
        smlal2 v24.4S, v19.8H, v0.8H         // ............................................................................*.....
        uzp1 v19.8H, v3.8H, v17.8H           // ......e...........................................................................
        uzp1 v26.8H, v18.8H, v13.8H          // ....................................................................*.............
        zip2 v14.8H, v7.8H, v15.8H           // ...............................................................................l..
        mul v23.8H, v26.8H, v2.8H            // .....................................................................*............
        uzp2 v15.8H, v29.8H, v24.8H          // .............................................................................*....
        smull2 v24.4S, v31.8H, v21.8H        // ..............e...................................................................
        str q14, [x0, #16]                   // .................................................................................l
        ldr q3, [x7, #16]                    // ...................................e..............................................
        ldr q6, [x8, #16]                    // .......................................e..........................................
        ldr q8, [x10], #32                   // ...................................................e..............................
        ldr q26, [x10, #-16]                 // ....................................................e.............................
        ld1 {v22.8H}, [x12], #16             // ...........................................................e......................
        uzp1 v30.8H, v8.8H, v26.8H           // .....................................................e............................
        uzp2 v11.8H, v8.8H, v26.8H           // ......................................................e...........................
        ldr q8, [x4], #32                    // .................e................................................................
        ldr q26, [x4, #-16]                  // ..................e...............................................................
        ldr q4, [x7], #32                    // ..................................e...............................................
        uzp1 v20.8H, v8.8H, v26.8H           // ...................e..............................................................
        uzp2 v26.8H, v8.8H, v26.8H           // ....................e.............................................................
        ld1 {v8.8H}, [x6], #16               // .........................e........................................................
        uzp1 v9.8H, v4.8H, v3.8H             // ....................................e.............................................
        ldr q25, [x11, #16]                  // ........................................................e.........................
        ldr q29, [x11], #32                  // .......................................................e..........................
        ld1 {v12.8H}, [x9], #16              // ..........................................e.......................................
        ldr q14, [x8], #32                   // ......................................e...........................................
        smlal2 v24.4S, v16.8H, v19.8H        // ................e.................................................................
        smlal2 v13.4S, v23.8H, v0.8H         // .......................................................................*..........
        smlal v18.4S, v23.4H, v0.4H          // ......................................................................*...........
        ld1 {v23.8H}, [x3], #16              // ........e.........................................................................
        smlal2 v24.4S, v20.8H, v27.8H        // ...............................e..................................................
        uzp2 v7.8H, v18.8H, v13.8H           // ........................................................................*.........
        uzp1 v10.8H, v29.8H, v25.8H          // .........................................................e........................
        str q5, [x0], #32                    // ................................................................................l.
        zip1 v5.8H, v7.8H, v15.8H            // ..............................................................................*...

                                             // ----------------------------------------------------------------------------------------------------------------- new position ------------------------------------------------------------------------------------------------------------------>
                                             // 0                        25                       50                       75                       100                      125                      150                      175                      200                      225
                                             // |------------------------|------------------------|------------------------|------------------------|------------------------|------------------------|------------------------|------------------------|------------------------|----------------
        // ldr q12, [x1], #32                // ....................................e..........................................'......................................~..........................................'......................................~.........................................
        // ldr q13, [x1, #-16]               // ................e..............................................................'..................~..............................................................'..................~.............................................................
        // uzp1 v3.8h, v12.8h, v13.8h        // ........................................e......................................'..........................................~......................................'..........................................~.....................................
        // uzp2 v4.8h, v12.8h, v13.8h        // .........................................e.....................................'...........................................~.....................................'...........................................~....................................
        // ldr q12, [x2], #32                // e..............................................................................'..~..............................................................................'..~.............................................................................
        // ldr q13, [x2, #-16]               // ...e...........................................................................'.....~...........................................................................'.....~..........................................................................
        // uzp1 v5.8h, v12.8h, v13.8h        // .............................................e.................................'...............................................~.................................'...............................................~................................
        // uzp2 v6.8h, v12.8h, v13.8h        // ..........................................e....................................'............................................~....................................'............................................~...................................
        // ld1 {v7.8h}, [x3], #16            // .........................................................................e.....'...........................................................................~.....'...........................................................................~....
        // smull  v8.4s, v3.4h, v5.4h        // ....~..........................................................................'......*..........................................................................'......~.........................................................................
        // smull2 v10.4s, v3.8h, v5.8h       // ...............................................................................'.*...............................................................................'.~..............................................................................
        // smlal  v8.4s, v4.4h, v7.4h        // ...........~...................................................................'.............*...................................................................'.............~..................................................................
        // smlal2 v10.4s, v4.8h, v7.8h       // ..~............................................................................'....*............................................................................'....~...........................................................................
        // smull  v9.4s, v3.4h, v6.4h        // ......~........................................................................'........*........................................................................'........~.......................................................................
        // smull2 v11.4s, v3.8h, v6.8h       // ..................................................e............................'....................................................~............................'....................................................~...........................
        // smlal  v9.4s, v4.4h, v5.4h        // .........~.....................................................................'...........*.....................................................................'...........~....................................................................
        // smlal2 v11.4s, v4.8h, v5.8h       // ......................................................................e........'........................................................................~........'........................................................................~.......
        // ldr q12, [x4], #32                // ...........................................................e...................'.............................................................~...................'.............................................................~..................
        // ldr q13, [x4, #-16]               // ............................................................e..................'..............................................................~..................'..............................................................~.................
        // uzp1 v3.8h, v12.8h, v13.8h        // ..............................................................e................'................................................................~................'................................................................~...............
        // uzp2 v4.8h, v12.8h, v13.8h        // ...............................................................e...............'.................................................................~...............'.................................................................~..............
        // ldr q12, [x5], #32                // .......e.......................................................................'.........~.......................................................................'.........~......................................................................
        // ldr q13, [x5, #-16]               // ..........e....................................................................'............~....................................................................'............~...................................................................
        // uzp1 v5.8h, v12.8h, v13.8h        // ......................................e........................................'........................................~........................................'........................................~.......................................
        // uzp2 v6.8h, v12.8h, v13.8h        // ..............e................................................................'................~................................................................'................~...............................................................
        // ld1 {v7.8h}, [x6], #16            // ................................................................e..............'..................................................................~..............'..................................................................~.............
        // smlal  v8.4s, v3.4h, v5.4h        // ...............~...............................................................'.................*...............................................................'.................~..............................................................
        // smlal2 v10.4s, v3.8h, v5.8h       // .....~.........................................................................'.......*.........................................................................'.......~........................................................................
        // smlal  v8.4s, v4.4h, v7.4h        // ..................~............................................................'....................*............................................................'....................~...........................................................
        // smlal2 v10.4s, v4.8h, v7.8h       // ........~......................................................................'..........*......................................................................'..........~.....................................................................
        // smlal  v9.4s, v3.4h, v6.4h        // ............~..................................................................'..............*..................................................................'..............~.................................................................
        // smlal2 v11.4s, v3.8h, v6.8h       // ..........................................................................e....'............................................................................~....'............................................................................~...
        // smlal  v9.4s, v4.4h, v5.4h        // .................~.............................................................'...................*.............................................................'...................~............................................................
        // smlal2 v11.4s, v4.8h, v5.8h       // ...............................................................................*.................................................................................~................................................................................
        // ldr q12, [x7], #32                // .............................................................e.................'...............................................................~.................'...............................................................~................
        // ldr q13, [x7, #-16]               // ....................................................e..........................'......................................................~..........................'......................................................~.........................
        // uzp1 v3.8h, v12.8h, v13.8h        // .................................................................e.............'...................................................................~.............'...................................................................~............
        // uzp2 v4.8h, v12.8h, v13.8h        // ...............................................................................'*................................................................................'~...............................................................................
        // ldr q12, [x8], #32                // .....................................................................e.........'.......................................................................~.........'.......................................................................~........
        // ldr q13, [x8, #-16]               // .....................................................e.........................'.......................................................~.........................'.......................................................~........................
        // uzp1 v5.8h, v12.8h, v13.8h        // .............~.................................................................'...............*.................................................................'...............~................................................................
        // uzp2 v6.8h, v12.8h, v13.8h        // ...................~...........................................................'.....................*...........................................................'.....................~..........................................................
        // ld1 {v7.8h}, [x9], #16            // ....................................................................e..........'......................................................................~..........'......................................................................~.........
        // smlal  v8.4s, v3.4h, v5.4h        // .......................~.......................................................'.........................*.......................................................'.........................~......................................................
        // smlal2 v10.4s, v3.8h, v5.8h       // ....................~..........................................................'......................*..........................................................'......................~.........................................................
        // smlal  v8.4s, v4.4h, v7.4h        // ...........................~...................................................'.............................*...................................................'.............................~..................................................
        // smlal2 v10.4s, v4.8h, v7.8h       // ........................~......................................................'..........................*......................................................'..........................~.....................................................
        // smlal  v9.4s, v3.4h, v6.4h        // ......................~........................................................'........................*........................................................'........................~.......................................................
        // smlal2 v11.4s, v3.8h, v6.8h       // .....................~.........................................................'.......................*.........................................................'.......................~........................................................
        // smlal  v9.4s, v4.4h, v5.4h        // ..........................~....................................................'............................*....................................................'............................~...................................................
        // smlal2 v11.4s, v4.8h, v5.8h       // .........................~.....................................................'...........................*.....................................................'...........................~....................................................
        // ldr q12, [x10], #32               // ......................................................e........................'........................................................~........................'........................................................~.......................
        // ldr q13, [x10, #-16]              // .......................................................e.......................'.........................................................~.......................'.........................................................~......................
        // uzp1 v3.8h, v12.8h, v13.8h        // .........................................................e.....................'...........................................................~.....................'...........................................................~....................
        // uzp2 v4.8h, v12.8h, v13.8h        // ..........................................................e....................'............................................................~....................'............................................................~...................
        // ldr q12, [x11], #32               // ...................................................................e...........'.....................................................................~...........'.....................................................................~..........
        // ldr q13, [x11, #-16]              // ..................................................................e............'....................................................................~............'....................................................................~...........
        // uzp1 v5.8h, v12.8h, v13.8h        // ............................................................................e..'..............................................................................~..'..............................................................................~.
        // uzp2 v6.8h, v12.8h, v13.8h        // .~.............................................................................'...*.............................................................................'...~............................................................................
        // ld1 {v7.8h}, [x12], #16           // ........................................................e......................'..........................................................~......................'..........................................................~.....................
        // smlal  v8.4s, v3.4h, v5.4h        // ...............................~...............................................'.................................*...............................................'.................................~..............................................
        // smlal2 v10.4s, v3.8h, v5.8h       // ............................~..................................................'..............................*..................................................'..............................~.................................................
        // smlal  v8.4s, v4.4h, v7.4h        // ...................................~...........................................'.....................................*...........................................'.....................................~..........................................
        // smlal2 v10.4s, v4.8h, v7.8h       // ................................~..............................................'..................................*..............................................'..................................~.............................................
        // smlal  v9.4s, v3.4h, v6.4h        // ..............................~................................................'................................*................................................'................................~...............................................
        // smlal2 v11.4s, v3.8h, v6.8h       // .............................~.................................................'...............................*.................................................'...............................~................................................
        // smlal  v9.4s, v4.4h, v5.4h        // ..................................~............................................'....................................*............................................'....................................~...........................................
        // smlal2 v11.4s, v4.8h, v5.8h       // .................................~.............................................'...................................*.............................................'...................................~............................................
        // uzp1   v28.8h, v8.8h, v10.8h      // ..............................................~................................'................................................*................................'................................................~...............................
        // mul    v28.8h, v28.8h, v2.8h      // ................................................~..............................'..................................................*..............................'..................................................~.............................
        // smlal  v8.4s, v28.4h, v0.4h       // ........................................................................~......'..........................................................................*......'..........................................................................~.....
        // smlal2 v10.4s, v28.8h, v0.8h      // .......................................................................~.......'.........................................................................*.......'.........................................................................~......
        // uzp2   v26.8h, v8.8h, v10.8h      // ...........................................................................~...'.............................................................................*...'.............................................................................~..
        // uzp1   v28.8h, v9.8h, v11.8h      // .....................................~.........................................'.......................................*.........................................'.......................................~........................................
        // mul    v28.8h, v28.8h, v2.8h      // .......................................~.......................................'.........................................*.......................................'.........................................~......................................
        // smlal  v9.4s, v28.4h, v0.4h       // ...........................................~...................................'.............................................*...................................'.............................................~..................................
        // smlal2 v11.4s, v28.8h, v0.8h      // ............................................~..................................'..............................................*..................................'..............................................~.................................
        // uzp2   v27.8h, v9.8h, v11.8h      // .................................................~.............................'...................................................*.............................'...................................................~............................
        // zip1 v12.8h, v26.8h, v27.8h       // ..............................................................................~'................................................................................*'................................................................................
        // zip2 v13.8h, v26.8h, v27.8h       // ...............................................~...............................'.................................................~...............................'.................................................l..............................
        // str q12, [x0], #32                // .............................................................................~.'...............................................................................~.'...............................................................................l
        // str q13, [x0, #-16]               // ...................................................~...........................'.....................................................~...........................'.....................................................l..........................

        sub count, count, #1
        cbnz count, polyvec_basemul_acc_montgomery_cached_asm_k4_opt_loop

                                             // Instructions:    50
                                             // Expected cycles: 56
                                             // Expected IPC:    0.89
                                             //
                                             // Cycle bound:     56.0
                                             // IPC bound:       0.89
                                             //
                                             // Wall time:     4.16s
                                             // User time:     4.16s
                                             //
                                             // --------------- original position --------------->
                                             // 0                        25
                                             // |------------------------|
        smull2 v17.4S, v31.8H, v19.8H        // ..*...............................................
        uzp2 v1.8H, v14.8H, v6.8H            // ................*.................................
        smull v18.4S, v31.4H, v21.4H         // .......*..........................................
        smlal2 v24.4S, v26.8H, v28.8H        // *.................................................
        smlal2 v17.4S, v16.8H, v23.8H        // ....*.............................................
        smull v21.4S, v31.4H, v19.4H         // .....*............................................
        smlal v18.4S, v16.4H, v19.4H         // .........*........................................
        uzp2 v31.8H, v4.8H, v3.8H            // .*................................................
        uzp1 v3.8H, v14.8H, v6.8H            // ............*.....................................
        smlal v21.4S, v16.4H, v23.4H         // ..........*.......................................
        smlal v18.4S, v20.4H, v27.4H         // ...........*......................................
        uzp2 v14.8H, v29.8H, v25.8H          // ...*..............................................
        smlal2 v17.4S, v20.8H, v28.8H        // ......*...........................................
        smlal v21.4S, v20.4H, v28.4H         // .............*....................................
        smlal v18.4S, v26.4H, v28.4H         // ..............*...................................
        smlal2 v24.4S, v9.8H, v1.8H          // ..................*...............................
        smlal2 v17.4S, v26.8H, v8.8H         // ........*.........................................
        smlal v21.4S, v26.4H, v8.4H          // ...............*..................................
        smlal v18.4S, v9.4H, v1.4H           // ...................*..............................
        smlal2 v24.4S, v31.8H, v3.8H         // ......................*...........................
        smlal2 v17.4S, v9.8H, v3.8H          // .................*................................
        smlal v21.4S, v9.4H, v3.4H           // ....................*.............................
        smlal v18.4S, v31.4H, v3.4H          // .......................*..........................
        smlal2 v24.4S, v30.8H, v14.8H        // ..........................*.......................
        smlal2 v17.4S, v31.8H, v12.8H        // .....................*............................
        smlal v21.4S, v31.4H, v12.4H         // ........................*.........................
        smlal v18.4S, v30.4H, v14.4H         // ...........................*......................
        smlal2 v24.4S, v11.8H, v10.8H        // ..............................*...................
        smlal2 v17.4S, v30.8H, v10.8H        // .........................*........................
        smlal v21.4S, v30.4H, v10.4H         // ............................*.....................
        smlal v18.4S, v11.4H, v10.4H         // ...............................*..................
        zip2 v19.8H, v7.8H, v15.8H           // ......................................*...........
        smlal2 v17.4S, v11.8H, v22.8H        // .............................*....................
        smlal v21.4S, v11.4H, v22.4H         // ................................*.................
        uzp1 v23.8H, v18.8H, v24.8H          // .................................*................
        str q19, [x0, #16]                   // .........................................*........
        mul v19.8H, v23.8H, v2.8H            // ..................................*...............
        uzp1 v23.8H, v21.8H, v17.8H          // .....................................*............
        str q5, [x0], #32                    // .............................................*....
        mul v26.8H, v23.8H, v2.8H            // .......................................*..........
        smlal v18.4S, v19.4H, v0.4H          // ...................................*..............
        smlal2 v24.4S, v19.8H, v0.8H         // ....................................*.............
        smlal v21.4S, v26.4H, v0.4H          // ...........................................*......
        smlal2 v17.4S, v26.8H, v0.8H         // ..........................................*.......
        uzp2 v13.8H, v18.8H, v24.8H          // ........................................*.........
        uzp2 v19.8H, v21.8H, v17.8H          // ............................................*.....
        zip1 v23.8H, v19.8H, v13.8H          // ..............................................*...
        zip2 v19.8H, v19.8H, v13.8H          // ...............................................*..
        str q23, [x0], #32                   // .................................................*
        str q19, [x0, #-16]                  // ................................................*.

                                              // ----------------- new position ------------------>
                                              // 0                        25
                                              // |------------------------|------------------------
        // smlal2 v24.4S, v26.8H, v28.8H      // ...*..............................................
        // uzp2 v4.8H, v4.8H, v3.8H           // .......*..........................................
        // smull2 v13.4S, v31.8H, v19.8H      // *.................................................
        // uzp2 v1.8H, v29.8H, v25.8H         // ...........*......................................
        // smlal2 v13.4S, v16.8H, v23.8H      // ....*.............................................
        // smull v18.4S, v31.4H, v19.4H       // .....*............................................
        // smlal2 v13.4S, v20.8H, v28.8H      // ............*.....................................
        // smull v29.4S, v31.4H, v21.4H       // ..*...............................................
        // smlal2 v13.4S, v26.8H, v8.8H       // ................*.................................
        // smlal v29.4S, v16.4H, v19.4H       // ......*...........................................
        // smlal v18.4S, v16.4H, v23.4H       // .........*........................................
        // smlal v29.4S, v20.4H, v27.4H       // ..........*.......................................
        // uzp1 v31.8H, v14.8H, v6.8H         // ........*.........................................
        // smlal v18.4S, v20.4H, v28.4H       // .............*....................................
        // smlal v29.4S, v26.4H, v28.4H       // ..............*...................................
        // smlal v18.4S, v26.4H, v8.4H        // .................*................................
        // uzp2 v26.8H, v14.8H, v6.8H         // .*................................................
        // smlal2 v13.4S, v9.8H, v31.8H       // ....................*.............................
        // smlal2 v24.4S, v9.8H, v26.8H       // ...............*..................................
        // smlal v29.4S, v9.4H, v26.4H        // ..................*...............................
        // smlal v18.4S, v9.4H, v31.4H        // .....................*............................
        // smlal2 v13.4S, v4.8H, v12.8H       // ........................*.........................
        // smlal2 v24.4S, v4.8H, v31.8H       // ...................*..............................
        // smlal v29.4S, v4.4H, v31.4H        // ......................*...........................
        // smlal v18.4S, v4.4H, v12.4H        // .........................*........................
        // smlal2 v13.4S, v30.8H, v10.8H      // ............................*.....................
        // smlal2 v24.4S, v30.8H, v1.8H       // .......................*..........................
        // smlal v29.4S, v30.4H, v1.4H        // ..........................*.......................
        // smlal v18.4S, v30.4H, v10.4H       // .............................*....................
        // smlal2 v13.4S, v11.8H, v22.8H      // ................................*.................
        // smlal2 v24.4S, v11.8H, v10.8H      // ...........................*......................
        // smlal v29.4S, v11.4H, v10.4H       // ..............................*...................
        // smlal v18.4S, v11.4H, v22.4H       // .................................*................
        // uzp1 v31.8H, v29.8H, v24.8H        // ..................................*...............
        // mul v19.8H, v31.8H, v2.8H          // ....................................*.............
        // smlal v29.4S, v19.4H, v0.4H        // ........................................*.........
        // smlal2 v24.4S, v19.8H, v0.8H       // .........................................*........
        // uzp1 v26.8H, v18.8H, v13.8H        // .....................................*............
        // zip2 v14.8H, v7.8H, v15.8H         // ...............................*..................
        // mul v23.8H, v26.8H, v2.8H          // .......................................*..........
        // uzp2 v15.8H, v29.8H, v24.8H        // ............................................*.....
        // str q14, [x0, #16]                 // ...................................*..............
        // smlal2 v13.4S, v23.8H, v0.8H       // ...........................................*......
        // smlal v18.4S, v23.4H, v0.4H        // ..........................................*.......
        // uzp2 v7.8H, v18.8H, v13.8H         // .............................................*....
        // str q5, [x0], #32                  // ......................................*...........
        // zip1 v5.8H, v7.8H, v15.8H          // ..............................................*...
        // zip2 v14.8H, v7.8H, v15.8H         // ...............................................*..
        // str q14, [x0, #16]                 // .................................................*
        // str q5, [x0], #32                  // ................................................*.


        pop_stack
        ret

/****************** REGISTER DEALLOCATIONS *******************/
    .unreq out
    .unreq a0_ptr
    .unreq b0_ptr
    .unreq b0_cache_ptr
    .unreq a1_ptr
    .unreq b1_ptr
    .unreq b1_cache_ptr
    .unreq a2_ptr
    .unreq b2_ptr
    .unreq b2_cache_ptr
    .unreq a3_ptr
    .unreq b3_ptr
    .unreq b3_cache_ptr
    .unreq count
    .unreq modulus
    .unreq modulus_twisted
    .unreq wtmp
    .unreq aa0
    .unreq aa1
    .unreq bb0
    .unreq bb1
    .unreq bb1t
    .unreq res0l
    .unreq res1l
    .unreq res0h
    .unreq res1h
    .unreq tmp0
    .unreq tmp1
    .unreq q_tmp0
    .unreq q_tmp1
    .unreq out0
    .unreq out1
    .unreq t0

/* simpasm: footer-start */
#endif /* defined(MLKEM_NATIVE_ARITH_BACKEND_AARCH64_OPT) && \
          !defined(MLKEM_NATIVE_MULTILEVEL_BUILD_NO_SHARED) && \
          (defined(MLKEM_NATIVE_MULTILEVEL_BUILD_WITH_SHARED) || MLKEM_K == 4) */
